"""Next-Generation CAREamics DataModule."""

import copy
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import Any, Literal, Union, overload

import numpy as np
import pytorch_lightning as L
from numpy.typing import NDArray
from torch.utils.data import DataLoader, Sampler
from torch.utils.data._utils.collate import default_collate

from careamics.config.data.ng_data_config import NGDataConfig
from careamics.config.support import SupportedData
from careamics.dataset_ng.factory import create_dataset
from careamics.dataset_ng.grouped_index_sampler import GroupedIndexSampler
from careamics.dataset_ng.image_stack_loader import ImageStackLoader
from careamics.lightning.dataset_ng.data_module_utils import initialize_data_pair
from careamics.utils import get_logger

logger = get_logger(__name__)

ItemType = Union[Path, str, NDArray[Any]]
"""Type of input items passed to the dataset."""

InputType = Union[ItemType, Sequence[ItemType], None]
"""Type of input data passed to the dataset."""


class CareamicsDataModule(L.LightningDataModule):
    """Data module for Careamics dataset.

    Parameters
    ----------
    data_config : DataConfig
        Pydantic model for CAREamics data configuration.
    train_data : Optional[InputType]
        Training data, can be a path to a folder, a list of paths, or a numpy array.
    train_data_target : Optional[InputType]
        Training data target, can be a path to a folder,
        a list of paths, or a numpy array.
    train_data_mask : InputType (when filtering is needed)
        Training data mask, can be a path to a folder,
        a list of paths, or a numpy array. Used for coordinate filtering.
        Only required when using coordinate-based patch filtering.
    val_data : Optional[InputType]
        Validation data, can be a path to a folder,
        a list of paths, or a numpy array.
    val_data_target : Optional[InputType]
        Validation data target, can be a path to a folder,
        a list of paths, or a numpy array.
    pred_data : Optional[InputType]
        Prediction data, can be a path to a folder, a list of paths,
        or a numpy array.
    pred_data_target : Optional[InputType]
        Prediction data target, can be a path to a folder,
        a list of paths, or a numpy array.
    read_source_func : Optional[Callable], default=None
        Function to read the source data. Only used for `custom`
        data type (see DataModel).
    read_kwargs : Optional[dict[str, Any]]
        The kwargs for the read source function.
    image_stack_loader : Optional[ImageStackLoader]
        The image stack loader.
    image_stack_loader_kwargs : Optional[dict[str, Any]]
        The image stack loader kwargs.
    extension_filter : str, default=""
        Filter for file extensions. Only used for `custom` data types
        (see DataModel).
    val_percentage : Optional[float]
        Percentage of the training data to use for validation. Only
        used if `val_data` is None.
    val_minimum_split : int, default=5
        Minimum number of patches or files to split from the training data for
        validation. Only used if `val_data` is None.


    Attributes
    ----------
    config : DataConfig
        Pydantic model for CAREamics data configuration.
    data_type : str
        Type of data, one of SupportedData.
    batch_size : int
        Batch size for the dataloaders.
    extension_filter : str
        Filter for file extensions, by default "".
    read_source_func : Optional[Callable], default=None
        Function to read the source data.
    read_kwargs : Optional[dict[str, Any]], default=None
        The kwargs for the read source function.
    val_percentage : Optional[float]
        Percentage of the training data to use for validation.
    val_minimum_split : int, default=5
        Minimum number of patches or files to split from the training data for
        validation.
    train_data : Optional[Any]
        Training data, can be a path to a folder, a list of paths, or a numpy array.
    train_data_target : Optional[Any]
        Training data target, can be a path to a folder, a list of paths, or a numpy
        array.
    train_data_mask : Optional[Any]
        Training data mask, can be a path to a folder, a list of paths, or a numpy
        array.
    val_data : Optional[Any]
        Validation data, can be a path to a folder, a list of paths, or a numpy array.
    val_data_target : Optional[Any]
        Validation data target, can be a path to a folder, a list of paths, or a numpy
        array.
    pred_data : Optional[Any]
        Prediction data, can be a path to a folder, a list of paths, or a numpy array.
    pred_data_target : Optional[Any]
        Prediction data target, can be a path to a folder, a list of paths, or a numpy
        array.

    Raises
    ------
    ValueError
        If at least one of train_data, val_data or pred_data is not provided.
    ValueError
        If input and target data types are not consistent.
    """

    # standard use (no mask)
    @overload
    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: InputType | None = None,
        train_data_target: InputType | None = None,
        val_data: InputType | None = None,
        val_data_target: InputType | None = None,
        pred_data: InputType | None = None,
        pred_data_target: InputType | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
    ) -> None: ...

    # with training mask for filtering
    @overload
    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: InputType | None = None,
        train_data_target: InputType | None = None,
        train_data_mask: InputType,
        val_data: InputType | None = None,
        val_data_target: InputType | None = None,
        pred_data: InputType | None = None,
        pred_data_target: InputType | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
    ) -> None: ...

    # custom read function (no mask)
    @overload
    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: InputType | None = None,
        train_data_target: InputType | None = None,
        val_data: InputType | None = None,
        val_data_target: InputType | None = None,
        pred_data: InputType | None = None,
        pred_data_target: InputType | None = None,
        read_source_func: Callable,
        read_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
    ) -> None: ...

    # custom read function with training mask
    @overload
    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: InputType | None = None,
        train_data_target: InputType | None = None,
        train_data_mask: InputType,
        val_data: InputType | None = None,
        val_data_target: InputType | None = None,
        pred_data: InputType | None = None,
        pred_data_target: InputType | None = None,
        read_source_func: Callable,
        read_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
    ) -> None: ...

    # image stack loader (no mask)
    @overload
    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: Any | None = None,
        train_data_target: Any | None = None,
        val_data: Any | None = None,
        val_data_target: Any | None = None,
        pred_data: Any | None = None,
        pred_data_target: Any | None = None,
        image_stack_loader: ImageStackLoader,
        image_stack_loader_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
    ) -> None: ...

    # image stack loader with training mask
    @overload
    def __init__(
        self,
        data_config: NGDataConfig,
        *,
        train_data: Any | None = None,
        train_data_target: Any | None = None,
        train_data_mask: Any,
        val_data: Any | None = None,
        val_data_target: Any | None = None,
        pred_data: Any | None = None,
        pred_data_target: Any | None = None,
        image_stack_loader: ImageStackLoader,
        image_stack_loader_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
    ) -> None: ...

    def __init__(
        self,
        data_config: NGDataConfig | dict[str, Any],
        *,
        train_data: Any | None = None,
        train_data_target: Any | None = None,
        train_data_mask: Any | None = None,
        val_data: Any | None = None,
        val_data_target: Any | None = None,
        pred_data: Any | None = None,
        pred_data_target: Any | None = None,
        read_source_func: Callable | None = None,
        read_kwargs: dict[str, Any] | None = None,
        image_stack_loader: ImageStackLoader | None = None,
        image_stack_loader_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
        val_percentage: float | None = None,
        val_minimum_split: int = 5,
    ) -> None:
        """
        Data module for Careamics dataset initialization.

        Create a lightning datamodule that handles creating datasets for training,
        validation, and prediction.

        Parameters
        ----------
        data_config : NGDataConfig
            Pydantic model for CAREamics data configuration.
        train_data : Optional[InputType]
            Training data, can be a path to a folder, a list of paths, or a numpy array.
        train_data_target : Optional[InputType]
            Training data target, can be a path to a folder,
            a list of paths, or a numpy array.
        train_data_mask : InputType (when filtering is needed)
            Training data mask, can be a path to a folder,
            a list of paths, or a numpy array. Used for coordinate filtering.
            Only required when using coordinate-based patch filtering.
        val_data : Optional[InputType]
            Validation data, can be a path to a folder,
            a list of paths, or a numpy array.
        val_data_target : Optional[InputType]
            Validation data target, can be a path to a folder,
            a list of paths, or a numpy array.
        pred_data : Optional[InputType]
            Prediction data, can be a path to a folder, a list of paths,
            or a numpy array.
        pred_data_target : Optional[InputType]
            Prediction data target, can be a path to a folder,
            a list of paths, or a numpy array.
        read_source_func : Optional[Callable]
            Function to read the source data, by default None. Only used for `custom`
            data type (see DataModel).
        read_kwargs : Optional[dict[str, Any]]
            The kwargs for the read source function.
        image_stack_loader : Optional[ImageStackLoader]
            The image stack loader.
        image_stack_loader_kwargs : Optional[dict[str, Any]]
            The image stack loader kwargs.
        extension_filter : str
            Filter for file extensions, by default "". Only used for `custom` data types
            (see DataModel).
        val_percentage : Optional[float]
            Percentage of the training data to use for validation. Only
            used if `val_data` is None.
        val_minimum_split : int
            Minimum number of patches or files to split from the training data for
            validation, by default 5. Only used if `val_data` is None.
        """
        super().__init__()

        if train_data is None and val_data is None and pred_data is None:
            raise ValueError(
                "At least one of train_data, val_data or pred_data must be provided."
            )
        elif train_data is None != val_data is None:
            raise ValueError(
                "If one of train_data or val_data is provided, both must be provided."
            )

        if isinstance(data_config, NGDataConfig):
            self.config = data_config
        else:
            self.config = NGDataConfig.model_validate(data_config)
        self.save_hyperparameters(
            {"data_config": self.config.model_dump(mode="json")},
            ignore=[
                "train_data",
                "train_data_target",
                "train_data_mask",
                "val_data",
                "val_data_target",
                "pred_data",
                "pred_data_target",
                "read_source_func",
                "read_kwargs",
                "image_stack_loader",
                "image_stack_loader_kwargs",
                "extension_filter",
                "val_percentage",
                "val_minimum_split",
            ],
        )

        self.data_type: str = self.config.data_type
        self.batch_size: int = self.config.batch_size

        self.extension_filter: str = (
            extension_filter  # list_files pulls the correct ext
        )
        self.read_source_func = read_source_func
        self.read_kwargs = read_kwargs
        self.image_stack_loader = image_stack_loader
        self.image_stack_loader_kwargs = image_stack_loader_kwargs

        # TODO: implement the validation split logic
        self.val_percentage = val_percentage
        self.val_minimum_split = val_minimum_split
        if self.val_percentage is not None:
            raise NotImplementedError("Validation split is not implemented.")

        custom_loader = self.image_stack_loader is not None
        self.train_data, self.train_data_target = initialize_data_pair(
            self.data_type,
            train_data,
            train_data_target,
            extension_filter,
            custom_loader,
        )
        self.train_data_mask, _ = initialize_data_pair(
            self.data_type, train_data_mask, None, extension_filter, custom_loader
        )

        self.val_data, self.val_data_target = initialize_data_pair(
            self.data_type, val_data, val_data_target, extension_filter, custom_loader
        )

        # The pred_data_target can be needed to count metrics on the prediction
        self.pred_data, self.pred_data_target = initialize_data_pair(
            self.data_type, pred_data, pred_data_target, extension_filter, custom_loader
        )

    def setup(self, stage: str) -> None:
        """
        Setup datasets.

        Lightning hook that is called at the beginning of fit (train + validate),
        validate, test, or predict. Creates the datasets for a given stage.

        Parameters
        ----------
        stage : str
            The stage to set up datasets for.
            Is either 'fit', 'validate', 'test', or 'predict'.

        Raises
        ------
        NotImplementedError
            If stage is not one of "fit", "validate" or "predict".
        """
        if stage == "fit":
            if self.config.mode != "training":
                raise ValueError(
                    f"CAREamicsDataModule configured for {self.config.mode} cannot be "
                    f"used for training. Please create a new CareamicsDataModule with "
                    f"a configuration with mode='training'."
                )

            self.train_dataset = create_dataset(
                config=self.config,
                inputs=self.train_data,
                targets=self.train_data_target,
                masks=self.train_data_mask,
                read_func=self.read_source_func,
                read_kwargs=self.read_kwargs,
                image_stack_loader=self.image_stack_loader,
                image_stack_loader_kwargs=self.image_stack_loader_kwargs,
            )

            validation_config = self.config.convert_mode("validating")

            self.val_dataset = create_dataset(
                config=validation_config,
                inputs=self.val_data,
                targets=self.val_data_target,
                read_func=self.read_source_func,
                read_kwargs=self.read_kwargs,
                image_stack_loader=self.image_stack_loader,
                image_stack_loader_kwargs=self.image_stack_loader_kwargs,
            )
        elif stage == "validate":
            validation_config = self.config.convert_mode("validating")
            self.val_dataset = create_dataset(
                config=validation_config,
                inputs=self.val_data,
                targets=self.val_data_target,
                read_func=self.read_source_func,
                read_kwargs=self.read_kwargs,
                image_stack_loader=self.image_stack_loader,
                image_stack_loader_kwargs=self.image_stack_loader_kwargs,
            )
        elif stage == "predict":
            if self.config.mode == "validating":
                raise ValueError(
                    "CAREamicsDataModule configured for validating cannot be used for "
                    "prediction. Please create a new CareamicsDataModule with a "
                    "configuration with mode='predicting'."
                )

            self.predict_dataset = create_dataset(
                config=(
                    self.config.convert_mode("predicting")
                    if self.config.mode == "training"
                    else self.config
                ),
                inputs=self.pred_data,
                targets=self.pred_data_target,
                read_func=self.read_source_func,
                read_kwargs=self.read_kwargs,
                image_stack_loader=self.image_stack_loader,
                image_stack_loader_kwargs=self.image_stack_loader_kwargs,
            )
        else:
            raise NotImplementedError(f"Stage {stage} not implemented")

    def _sampler(self, dataset: Literal["train", "val", "predict"]) -> Sampler | None:
        sampler: GroupedIndexSampler | None
        rng = np.random.default_rng(self.config.seed)
        if not self.config.in_memory and self.config.data_type == SupportedData.TIFF:
            match dataset:
                case "train":
                    ds = self.train_dataset
                case "val":
                    ds = self.val_dataset
                case "predict":
                    ds = self.predict_dataset
                case _:
                    raise (
                        f"Unrecognized dataset '{dataset}', should be one of 'train', "
                        "'val' or 'predict'."
                    )
            sampler = GroupedIndexSampler.from_dataset(ds, rng=rng)
        else:
            sampler = None
        return sampler

    def train_dataloader(self) -> DataLoader:
        """
        Create a dataloader for training.

        Returns
        -------
        DataLoader
            Training dataloader.
        """
        sampler = self._sampler("train")
        dataloader_params = copy.deepcopy(self.config.train_dataloader_params)
        # have to remove shuffle with sampler because of torch error:
        #   ValueError: sampler option is mutually exclusive with shuffle
        # TODO: there might be other parameters mutually exclusive with sampler
        if (sampler is not None) and ("shuffle" in dataloader_params):
            del dataloader_params["shuffle"]
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            collate_fn=default_collate,
            sampler=sampler,
            **dataloader_params,
        )

    def val_dataloader(self) -> DataLoader:
        """
        Create a dataloader for validation.

        Returns
        -------
        DataLoader
            Validation dataloader.
        """
        sampler = self._sampler("val")
        dataloader_params = copy.deepcopy(self.config.val_dataloader_params)
        if (sampler is not None) and ("shuffle" in dataloader_params):
            del dataloader_params["shuffle"]
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            collate_fn=default_collate,
            sampler=sampler,
            **dataloader_params,
        )

    def predict_dataloader(self) -> DataLoader:
        """
        Create a dataloader for prediction.

        Returns
        -------
        DataLoader
            Prediction dataloader.
        """
        return DataLoader(
            self.predict_dataset,
            batch_size=self.batch_size,
            collate_fn=default_collate,
            **self.config.pred_dataloader_params,
        )



"""Noise2Void Lightning Module."""

from typing import Any, cast

import pytorch_lightning as L
import torch
from torch import nn
from torchmetrics import MetricCollection
from torchmetrics.image import PeakSignalNoiseRatio

from careamics.config import N2VAlgorithm, algorithm_factory
from careamics.dataset_ng.dataset import ImageRegionData
from careamics.losses import n2v_loss
from careamics.models.unet import UNet
from careamics.transforms import N2VManipulateTorch
from careamics.utils.logging import get_logger

from .module_utils import configure_optimizers, log_training_stats, log_validation_stats

logger = get_logger(__name__)


class N2VModule(L.LightningModule):
    """CAREamics PyTorch Lightning module for N2V algorithm.

    Parameters
    ----------
    algorithm_config : N2VAlgorithm or dict
        Configuration for the N2V algorithm, either as an N2VAlgorithm instance or a
        dictionary.
    """

    def __init__(self, algorithm_config: N2VAlgorithm | dict[str, Any]) -> None:
        """Instantiate N2VModule.

        Parameters
        ----------
        algorithm_config : N2VAlgorithm or dict
            Configuration for the N2V algorithm, either as an N2VAlgorithm instance or a
            dictionary.
        """
        super().__init__()

        if isinstance(algorithm_config, dict):
            config = algorithm_factory(algorithm_config)
        else:
            config = algorithm_config

        if not isinstance(config, N2VAlgorithm):
            raise TypeError("algorithm_config must be a N2VAlgorithm")

        self.save_hyperparameters({"algorithm_config": config.model_dump(mode="json")})
        self.config = config
        self.model: nn.Module = UNet(**self.config.model.model_dump())
        self.n2v_manipulate = N2VManipulateTorch(
            n2v_manipulate_config=self.config.n2v_config
        )
        self.loss_func = n2v_loss

        self.metrics = MetricCollection(PeakSignalNoiseRatio())

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass.

        Parameters
        ----------
        x : torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            Model output tensor.
        """
        return self.model(x)

    def training_step(
        self,
        batch: tuple[ImageRegionData] | tuple[ImageRegionData, ImageRegionData],
        batch_idx: int,
    ) -> torch.Tensor:
        """Training step for N2V model.

        Parameters
        ----------
        batch : ImageRegionData or (ImageRegionData, ImageRegionData)
            A tuple containing the input data and the target data.
        batch_idx : int
            The index of the current batch in the training loop.

        Returns
        -------
        torch.Tensor
            The loss value for the current training step.
        """
        x = batch[0]
        x_data = cast(torch.Tensor, x.data)
        x_masked, x_original, mask = self.n2v_manipulate(x_data)
        prediction = self.model(x_masked)
        loss = self.loss_func(prediction, x_original, mask)

        log_training_stats(self, loss, batch_size=x_data.shape[0])

        return loss

    def validation_step(
        self,
        batch: tuple[ImageRegionData] | tuple[ImageRegionData, ImageRegionData],
        batch_idx: int,
    ) -> None:
        """Validation step for N2V model.

        Parameters
        ----------
        batch : ImageRegionData or (ImageRegionData, ImageRegionData)
            A tuple containing the input data and the target data.
        batch_idx : int
            The index of the current batch in the validation loop.
        """
        x = batch[0]
        x_data = cast(torch.Tensor, x.data)
        x_masked, x_original, mask = self.n2v_manipulate(x_data)
        prediction = self.model(x_masked)
        val_loss = self.loss_func(prediction, x_original, mask)
        self.metrics(prediction, x_original)
        log_validation_stats(
            self, val_loss, batch_size=x_data.shape[0], metrics=self.metrics
        )

    def predict_step(
        self,
        batch: tuple[ImageRegionData] | tuple[ImageRegionData, ImageRegionData],
        batch_idx: int,
    ) -> ImageRegionData:
        """Prediction step for N2V model.

        Parameters
        ----------
        batch : ImageRegionData or (ImageRegionData, ImageRegionData)
            A tuple containing the input data and optionally the target data.
        batch_idx : int
            The index of the current batch in the prediction loop.

        Returns
        -------
        ImageRegionData
            The output batch containing the predictions.
        """
        x = batch[0]
        x_data = cast(torch.Tensor, x.data)
        # TODO: add TTA
        prediction = self.model(x_data)

        normalization = self._trainer.datamodule.predict_dataset.normalization  # type: ignore[union-attr]
        denormalized_output = normalization.denormalize(prediction).cpu().numpy()

        output_batch = ImageRegionData(
            data=denormalized_output,
            source=x.source,
            data_shape=x.data_shape,
            dtype=x.dtype,
            axes=x.axes,
            region_spec=x.region_spec,
            additional_metadata={},
        )
        return output_batch

    def configure_optimizers(self) -> dict[str, Any]:  # type: ignore[override]
        """Configure optimizer and learning rate scheduler.

        Returns
        -------
        dict[str, Any]
            A dictionary containing the optimizer and learning rate scheduler.
        """
        return configure_optimizers(
            model=self.model,
            optimizer_name=self.config.optimizer.name,
            optimizer_parameters=self.config.optimizer.parameters,
            lr_scheduler_name=self.config.lr_scheduler.name,
            lr_scheduler_parameters=self.config.lr_scheduler.parameters,
        )



from collections.abc import Sequence

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm

from careamics.config.data.normalization_config import (
    MeanStdConfig,
    MinMaxConfig,
    NoNormConfig,
    NormalizationConfig,
    QuantileConfig,
)
from careamics.dataset.dataset_utils.running_stats import WelfordStatistics
from careamics.dataset_ng.normalization.running_quantile import QuantileEstimator
from careamics.dataset_ng.patch_extractor import PatchExtractor
from careamics.dataset_ng.patching_strategies import PatchingStrategy


def _compute_mean_std(
    data_extractor: PatchExtractor,
    patching_strategy: PatchingStrategy,
    channels: Sequence[int] | None = None,
) -> tuple[list[float], list[float]]:
    """Compute mean and std from data."""
    image_stats = WelfordStatistics()
    n_patches = patching_strategy.n_patches

    for idx in tqdm(range(n_patches), desc="Computing mean/std statistics"):
        patch_spec = patching_strategy.get_patch_spec(idx)
        patch = data_extractor.extract_channel_patch(
            data_idx=patch_spec["data_idx"],
            sample_idx=patch_spec["sample_idx"],
            channels=channels,
            coords=patch_spec["coords"],
            patch_size=patch_spec["patch_size"],
        )
        image_stats.update(patch[None, ...], sample_idx=idx)

    means, stds = image_stats.finalize()
    return means.tolist(), stds.tolist()


def _compute_min_max(
    data_extractor: PatchExtractor,
    patching_strategy: PatchingStrategy,
    channels: Sequence[int] | None = None,
) -> tuple[list[float], list[float]]:
    """Compute min and max from data."""
    n_patches = patching_strategy.n_patches
    if n_patches == 0:
        raise ValueError("No patches found to compute min/max statistics.")

    first_spec = patching_strategy.get_patch_spec(0)
    first_patch = data_extractor.extract_patch(
        data_idx=first_spec["data_idx"],
        sample_idx=first_spec["sample_idx"],
        coords=first_spec["coords"],
        patch_size=first_spec["patch_size"],
    )
    axes = tuple(dim for dim in range(first_patch.ndim) if dim != 0)
    min_vals: NDArray = np.min(first_patch, axis=axes)
    max_vals: NDArray = np.max(first_patch, axis=axes)

    for idx in tqdm(range(1, n_patches), desc="Computing min/max statistics"):
        patch_spec = patching_strategy.get_patch_spec(idx)
        patch = data_extractor.extract_channel_patch(
            data_idx=patch_spec["data_idx"],
            sample_idx=patch_spec["sample_idx"],
            channels=channels,
            coords=patch_spec["coords"],
            patch_size=patch_spec["patch_size"],
        )
        current_mins = np.min(patch, axis=axes)
        current_maxes = np.max(patch, axis=axes)
        min_vals = np.minimum(min_vals, current_mins)
        max_vals = np.maximum(max_vals, current_maxes)

    return min_vals.tolist(), max_vals.tolist()


def _compute_quantiles(
    data_extractor: PatchExtractor,
    patching_strategy: PatchingStrategy,
    lower_quantiles: list[float],
    upper_quantiles: list[float],
    channels: Sequence[int] | None = None,
) -> tuple[list[float], list[float]]:
    """Compute quantile values from data."""
    estimator = QuantileEstimator(
        lower_quantiles=lower_quantiles,
        upper_quantiles=upper_quantiles,
    )

    n_patches = patching_strategy.n_patches
    for idx in tqdm(range(n_patches), desc="Computing quantile statistics"):
        patch_spec = patching_strategy.get_patch_spec(idx)
        patch = data_extractor.extract_channel_patch(
            data_idx=patch_spec["data_idx"],
            sample_idx=patch_spec["sample_idx"],
            channels=channels,
            coords=patch_spec["coords"],
            patch_size=patch_spec["patch_size"],
        )
        estimator.update(patch)

    lower_values, upper_values = estimator.finalize()
    return lower_values.tolist(), upper_values.tolist()


def resolve_normalization_config(
    norm_config: NormalizationConfig,
    patching_strategy: PatchingStrategy,
    input_extractor: PatchExtractor,
    target_extractor: PatchExtractor | None = None,
    channels: Sequence[int] | None = None,
) -> NormalizationConfig:
    """
    Resolve a normalization config by computing any missing statistics.

    If statistics are already provided in the config, they are preserved.
    If statistics are missing (None), they are computed from the data.

    Parameters
    ----------
    norm_config : NormalizationConfig
        The normalization configuration (may have missing statistics).
    patching_strategy : PatchingStrategy
        Strategy for iterating over patches.
    channels : list[int]
        Channels to compute statistics for.
    input_extractor : PatchExtractor
        Extractor for input data.
    target_extractor : PatchExtractor, optional
        Extractor for target data.

    Returns
    -------
    NormalizationConfig
        A resolved configuration with all statistics populated.
    """
    if isinstance(norm_config, NoNormConfig):
        return norm_config

    if isinstance(norm_config, MeanStdConfig):
        if norm_config.needs_computation():
            input_means, input_stds = _compute_mean_std(
                input_extractor, patching_strategy, channels
            )
            norm_config.set_input_stats(input_means, input_stds)

        if target_extractor is not None and norm_config.target_means is None:
            target_means, target_stds = _compute_mean_std(
                target_extractor, patching_strategy, channels
            )
            norm_config.set_target_stats(target_means, target_stds)

        return norm_config

    if isinstance(norm_config, MinMaxConfig):
        if norm_config.needs_computation():
            input_mins, input_maxes = _compute_min_max(
                input_extractor, patching_strategy, channels
            )
            norm_config.set_input_range(input_mins, input_maxes)

        if target_extractor is not None and norm_config.target_mins is None:
            target_mins, target_maxes = _compute_min_max(
                target_extractor, patching_strategy, channels
            )
            norm_config.set_target_range(target_mins, target_maxes)

        return norm_config

    if isinstance(norm_config, QuantileConfig):
        if norm_config.needs_computation():
            first_spec = patching_strategy.get_patch_spec(0)
            first_patch = input_extractor.extract_channel_patch(
                data_idx=first_spec["data_idx"],
                sample_idx=first_spec["sample_idx"],
                channels=channels,
                coords=first_spec["coords"],
                patch_size=first_spec["patch_size"],
            )
            n_channels = first_patch.shape[0]

            lower_levels = norm_config.get_lower_quantiles_for_channels(n_channels)
            upper_levels = norm_config.get_upper_quantiles_for_channels(n_channels)

            lower_values, upper_values = _compute_quantiles(
                input_extractor,
                patching_strategy,
                lower_levels,
                upper_levels,
                channels=channels,
            )
            norm_config.set_input_quantile_values(lower_values, upper_values)

        if (
            target_extractor is not None
            and norm_config.target_lower_quantile_values is None
        ):
            first_spec = patching_strategy.get_patch_spec(0)
            first_patch = target_extractor.extract_channel_patch(
                data_idx=first_spec["data_idx"],
                sample_idx=first_spec["sample_idx"],
                channels=channels,
                coords=first_spec["coords"],
                patch_size=first_spec["patch_size"],
            )
            n_channels = first_patch.shape[0]

            lower_levels = norm_config.get_lower_quantiles_for_channels(n_channels)
            upper_levels = norm_config.get_upper_quantiles_for_channels(n_channels)

            target_lower, target_upper = _compute_quantiles(
                target_extractor,
                patching_strategy,
                lower_levels,
                upper_levels,
                channels=channels,
            )
            norm_config.set_target_quantile_values(target_lower, target_upper)

        return norm_config

    raise ValueError(f"Unknown normalization config type: {type(norm_config)}")



"""CAREamics PyTorch Lightning modules."""

from .care_module import CAREModule
from .get_module import (
    CAREamicsModule,
    create_module,
    get_module_cls,
    load_module_from_checkpoint,
)
from .n2v_module import N2VModule

__all__ = [
    "CAREModule",
    "CAREamicsModule",
    "N2VModule",
    "create_module",
    "get_module_cls",
    "load_module_from_checkpoint",
]



from collections.abc import Sequence

from numpy.typing import NDArray

from ..image_stack import FileImageStack
from .patch_construction import PatchConstructor, default_patch_constr
from .patch_extractor import PatchExtractor


class LimitFilesPatchExtractor(PatchExtractor[FileImageStack]):
    """
    A patch extractor that limits the number of files that have their data loaded.

    This is useful for when not all of the data will fit into memory.
    """

    def __init__(
        self,
        image_stacks: Sequence[FileImageStack],
        patch_constructor: PatchConstructor = default_patch_constr,
    ):
        """
        Parameters
        ----------
        image_stacks: Sequence of `FileImageStack`
        """
        super().__init__(image_stacks, patch_constructor)
        self.loaded_stacks: list[int] = []

    def extract_channel_patch(
        self,
        data_idx: int,
        sample_idx: int,
        channels: Sequence[int] | None,
        coords: Sequence[int],
        patch_size: Sequence[int],
    ) -> NDArray:
        if data_idx not in self.loaded_stacks:
            # TODO: make maximum images loaded configurable?
            if len(self.loaded_stacks) >= 1:
                # get the idx that was added longest ago
                idx_to_close = self.loaded_stacks.pop(0)
                self.image_stacks[idx_to_close].close()

            self.image_stacks[data_idx].load()
            self.loaded_stacks.append(data_idx)

        return super().extract_channel_patch(
            data_idx, sample_idx, channels, coords, patch_size
        )



"""A package for the `PredictionWriterCallback` class and utilities."""

__all__ = [
    "CachedTiles",
    "PredictionWriterCallback",
    "WriteImage",
    "WriteStrategy",
    "WriteTilesZarr",
    "create_write_file_path",
    "create_write_strategy",
    "decollate_image_region_data",
    "select_write_extension",
    "select_write_func",
]

from .cached_tiles_strategy import CachedTiles
from .file_path_utils import create_write_file_path
from .prediction_writer_callback import (
    PredictionWriterCallback,
    decollate_image_region_data,
)
from .write_image_strategy import WriteImage
from .write_strategy import WriteStrategy
from .write_strategy_factory import (
    create_write_strategy,
    select_write_extension,
    select_write_func,
)
from .write_tiles_zarr_strategy import WriteTilesZarr



from functools import partial
from typing import Any

from typing_extensions import ParamSpec

from careamics.config.data.ng_data_config import NGDataConfig
from careamics.config.support import SupportedData
from careamics.file_io.read import ReadFunc

from .dataset import CareamicsDataset
from .image_stack import (
    GenericImageStack,
    ImageStack,
)
from .image_stack_loader import (
    ImageStackLoader,
    load_arrays,
    load_custom_file,
    load_czis,
    load_iter_tiff,
    load_tiffs,
    load_zarrs,
)
from .patch_extractor import LimitFilesPatchExtractor, PatchExtractor

P = ParamSpec("P")


# convenience function but should use `create_dataloader` function instead
# For lazy loading custom batch sampler also needs to be set.
def create_dataset(
    config: NGDataConfig,
    inputs: Any,
    targets: Any,
    masks: Any = None,
    read_func: ReadFunc | None = None,
    read_kwargs: dict[str, Any] | None = None,
    image_stack_loader: ImageStackLoader | None = None,
    image_stack_loader_kwargs: dict[str, Any] | None = None,
) -> CareamicsDataset[ImageStack]:
    """
    Convenience function to create the CAREamicsDataset.

    Parameters
    ----------
    config : DataConfig or InferenceConfig
        The data configuration.
    inputs : Any
        The input sources to the dataset.
    targets : Any, optional
        The target sources to the dataset.
    masks : Any, optional
        The mask sources used to filter patches.
    read_func : ReadFunc, optional
        A function that can that can be used to load custom data. This argument is
        ignored unless the `data_type` in the `config` is "custom".
    read_kwargs : dict of {str, Any}, optional
        Additional key-word arguments to pass to the `read_func`.
    image_stack_loader : ImageStackLoader, optional
        A function for custom image stack loading. This argument is ignored unless the
        `data_type` in the `config` is "custom".
    image_stack_loader_kwargs : {str, Any}, optional
        Additional key-word arguments to pass to the `image_stack_loader`.
    """
    image_stack_loader = select_image_stack_loader(
        data_type=SupportedData(config.data_type),
        in_memory=config.in_memory,
        read_func=read_func,
        read_kwargs=read_kwargs,
        image_stack_loader=image_stack_loader,
        image_stack_loader_kwargs=image_stack_loader_kwargs,
    )
    patch_extractor_type = select_patch_extractor_type(
        data_type=SupportedData(config.data_type), in_memory=config.in_memory
    )
    input_extractor = init_patch_extractor(
        patch_extractor_type, image_stack_loader, inputs, config.axes
    )
    if targets is not None:
        target_extractor = init_patch_extractor(
            patch_extractor_type, image_stack_loader, targets, config.axes
        )
    else:
        target_extractor = None
    if masks is not None:
        mask_extractor = init_patch_extractor(
            patch_extractor_type, image_stack_loader, masks, config.axes
        )
    else:
        mask_extractor = None
    return CareamicsDataset(
        data_config=config,
        input_extractor=input_extractor,
        target_extractor=target_extractor,
        mask_extractor=mask_extractor,
    )


def init_patch_extractor(
    patch_extractor: type[PatchExtractor],
    image_stack_loader: ImageStackLoader[..., GenericImageStack],
    source: Any,
    axes: str,
) -> PatchExtractor[GenericImageStack]:
    image_stacks = image_stack_loader(source, axes)
    return patch_extractor(image_stacks)


def select_patch_extractor_type(
    data_type: SupportedData,
    in_memory: bool,
) -> type[PatchExtractor]:
    """Select the appropriate PatchExtractor type based on data type and memory mode.

    If `in_memory` is True, or `data_type` is ZARR or CZI, the standard
    `PatchExtractor` is selected, otherwise the `LimitFilesPatchExtractor` will be used.

    Parameters
    ----------
    data_type : SupportedData
        The type of data being handled.
    in_memory : bool
        Indicates whether data is to be loaded into memory.

    Returns
    -------
    type[PatchExtractor]
        The selected PatchExtractor type.
    """
    if not in_memory and data_type in (SupportedData.TIFF, SupportedData.CUSTOM):
        return LimitFilesPatchExtractor
    else:
        return PatchExtractor


def select_image_stack_loader(
    data_type: SupportedData,
    in_memory: bool,
    read_func: ReadFunc | None = None,
    read_kwargs: dict[str, Any] | None = None,
    image_stack_loader: ImageStackLoader | None = None,
    image_stack_loader_kwargs: dict[str, Any] | None = None,
) -> ImageStackLoader:
    match data_type:
        case SupportedData.ARRAY:
            return load_arrays
        case SupportedData.TIFF:
            if in_memory:
                return load_tiffs
            else:
                return load_iter_tiff
        case SupportedData.CUSTOM:
            if (read_func is not None) and (image_stack_loader is None):
                read_kwargs = {} if read_kwargs is None else read_kwargs
                return partial(
                    load_custom_file, read_func=read_func, read_kwargs=read_kwargs
                )
            elif (read_func is None) and (image_stack_loader is not None):
                image_stack_loader_kwargs = (
                    {}
                    if image_stack_loader_kwargs is None
                    else image_stack_loader_kwargs
                )
                return partial(image_stack_loader, **image_stack_loader_kwargs)
            else:
                raise ValueError(
                    "Found `data_type='custom'` **one** of `read_func` or "
                    "`image_stack_loader` must be provided."
                )
        case SupportedData.ZARR:
            # TODO: in_memory or not
            return load_zarrs
        case SupportedData.CZI:
            # TODO: in_memory or not
            return load_czis
        case _:
            raise NotImplementedError(
                f"Selecting an image stack for data type '{data_type}' has not been "
                "implemented yet."
            )



import pytest

from careamics.config.algorithms import PN2VAlgorithm


def test_pn2v_requires_noise_model():
    """Check that PN2V algorithm fails without a noise model."""
    model = {
        "architecture": "UNet",
        "in_channels": 1,
        "num_classes": 1,
        "n2v2": False,
    }

    # Test that PN2V algorithm fails without noise model
    with pytest.raises(ValueError):
        PN2VAlgorithm(
            algorithm="pn2v",
            loss="n2v",
            model=model,
            # Missing noise_model parameter
        )



from collections.abc import Callable
from pathlib import Path
from typing import Any, Literal, TypedDict, Unpack

import torch
from numpy.typing import NDArray
import numpy as np
from pytorch_lightning import Callback, Trainer
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from pytorch_lightning.loggers import CSVLogger, TensorBoardLogger, WandbLogger

from careamics.lightning.dataset_ng.lightning_modules.get_module import CAREamicsModule

from .config import load_configuration_ng
from .config.ng_configs import N2VConfiguration
from .config.support import SupportedAlgorithm, SupportedLogger, SupportedData
from .file_io import WriteFunc
from .dataset.dataset_utils import reshape_array
from .model_io import export_to_bmz
from .lightning.callbacks import CareamicsCheckpointInfo, ProgressBarCallback
from .lightning.dataset_ng.callbacks.prediction_writer import PredictionWriterCallback
from .lightning.dataset_ng.lightning_modules import (
    CAREamicsModule,
    create_module,
    load_module_from_checkpoint,
)
from .utils import get_logger
from .utils.lightning_utils import read_csv_logger

logger = get_logger(__name__)

ExperimentLogger = TensorBoardLogger | WandbLogger | CSVLogger
Configuration = N2VConfiguration


class UserContext(TypedDict, total=False):
    work_dir: Path | str | None
    callbacks: list[Callback] | None
    enable_progress_bar: bool


class CAREamistV2:
    def __init__(
        self,
        config: Configuration | Path | None = None,
        *,
        checkpoint_path: Path | None = None,
        bmz_path: Path | None = None,
        **user_context: Unpack[UserContext],
    ):
        self.checkpoint_path = checkpoint_path
        self.work_dir = self._resolve_work_dir(user_context.get("work_dir"))
        self.config, self.model = self._load_model(config, checkpoint_path, bmz_path)

        enable_progress_bar = user_context.get("enable_progress_bar", True)
        self.config.training_config.lightning_trainer_config["enable_progress_bar"] = (
            enable_progress_bar
        )
        callbacks = user_context.get("callbacks", None)
        self.callbacks = self._define_callbacks(callbacks, self.config, self.work_dir)

        self.prediction_writer = PredictionWriterCallback(self.work_dir)
        self.prediction_writer.disable_writing(True)

        experiment_loggers = self._create_loggers(
            self.config.training_config.logger,
            self.config.experiment_name,
            self.work_dir,
        )

        self.trainer = Trainer(
            callbacks=[self.prediction_writer, *self.callbacks],
            default_root_dir=self.work_dir,
            logger=experiment_loggers,
            **self.config.training_config.lightning_trainer_config or {},
        )

    def _load_model(
        self,
        config: Configuration | Path | None,
        checkpoint_path: Path | None,
        bmz_path: Path | None,
    ) -> tuple[Configuration, CAREamicsModule]:
        n_inputs = sum(
            [config is not None, checkpoint_path is not None, bmz_path is not None]
        )
        if n_inputs != 1:
            raise ValueError(
                "Exactly one of `config`, `checkpoint_path`, or `bmz_path` must be provided."
            )
        if config is not None:
            return self._from_config(config)
        elif checkpoint_path is not None:
            return self._from_checkpoint(checkpoint_path)
        else:
            assert bmz_path is not None
            return self._from_bmz(bmz_path)

    @staticmethod
    def _from_config(
        config: Configuration | Path,
    ) -> tuple[Configuration, CAREamicsModule]:
        if isinstance(config, Path):
            config = load_configuration_ng(config)
        assert not isinstance(config, Path)

        model = create_module(config.algorithm_config)
        return config, model

    @staticmethod
    def _from_checkpoint(
        checkpoint_path: Path,
    ) -> tuple[Configuration, CAREamicsModule]:
        checkpoint: dict = torch.load(checkpoint_path, map_location="cpu")

        careamics_info = checkpoint.get("careamics_info", None)
        if careamics_info is None:
            raise ValueError(
                "Could not find CAREamics related information within the provided "
                "checkpoint. This means that it was saved without using the "
                "CAREamics callback `CareamicsCheckpointInfo`. "
                "Please use a checkpoint saved with CAREamics or initialize with a config instead."
            )

        try:
            algorithm_config: dict[str, Any] = checkpoint["hyper_parameters"][
                "algorithm_config"
            ]
        except (KeyError, IndexError) as e:
            raise ValueError(
                "Could not determine CAREamics supported algorithm from the provided "
                f"checkpoint at: {checkpoint_path!s}."
            ) from e

        data_hparams_key = checkpoint.get(
            "datamodule_hparams_name", "datamodule_hyper_parameters"
        )
        try:
            data_config: dict[str, Any] = checkpoint[data_hparams_key]["data_config"]
        except (KeyError, IndexError) as e:
            raise ValueError(
                "Could not determine the data configuration from the provided "
                f"checkpoint at: {checkpoint_path!s}."
            ) from e

        # TODO: will need to resolve this with type adapter once more configs are added
        config = Configuration.model_validate(
            {
                "algorithm_config": algorithm_config,
                "data_config": data_config,
                **careamics_info,
            }
        )

        module = load_module_from_checkpoint(checkpoint_path)
        return config, module

    @staticmethod
    def _from_bmz(
        bmz_path: Path,
    ) -> tuple[Configuration, CAREamicsModule]:
        raise NotImplementedError("Loading from BMZ is not implemented yet.")

    @staticmethod
    def _resolve_work_dir(work_dir: str | Path | None) -> Path:
        if work_dir is None:
            work_dir = Path.cwd().resolve()
            logger.warning(
                f"No working directory provided. Using current working directory: "
                f"{work_dir}."
            )
        else:
            work_dir = Path(work_dir).resolve()
        return work_dir

    @staticmethod
    def _define_callbacks(
        callbacks: list[Callback] | None,
        config: Configuration,
        work_dir: Path,
    ) -> list[Callback]:
        callbacks = [] if callbacks is None else callbacks
        for c in callbacks:
            if isinstance(c, (ModelCheckpoint, EarlyStopping)):
                raise ValueError(
                    "`ModelCheckpoint` and `EarlyStopping` callbacks are already "
                    "defined in CAREamics and should only be modified through the "
                    "training configuration (see TrainingConfig)."
                )

            if isinstance(c, (CareamicsCheckpointInfo, ProgressBarCallback)):
                raise ValueError(
                    "`CareamicsCheckpointInfo` and `ProgressBar` callbacks are defined "
                    "internally and should not be passed as callbacks."
                )

        internal_callbacks = [
            ModelCheckpoint(
                dirpath=work_dir / "checkpoints",
                filename=f"{config.experiment_name}_{{epoch:02d}}_step_{{step}}",
                **config.training_config.checkpoint_callback.model_dump(),
            ),
            CareamicsCheckpointInfo(
                config.version, config.experiment_name, config.training_config
            ),
        ]

        enable_progress_bar = config.training_config.lightning_trainer_config.get(
            "enable_progress_bar", True
        )
        if enable_progress_bar:
            internal_callbacks.append(ProgressBarCallback())

        if config.training_config.early_stopping_callback is not None:
            internal_callbacks.append(
                EarlyStopping(
                    **config.training_config.early_stopping_callback.model_dump()
                )
            )

        return internal_callbacks + callbacks

    @staticmethod
    def _create_loggers(
        logger: str | None, experiment_name: str, work_dir: Path
    ) -> list[ExperimentLogger]:
        csv_logger = CSVLogger(name=experiment_name, save_dir=work_dir / "csv_logs")

        if logger is not None:
            logger = SupportedLogger(logger)

        match logger:
            case SupportedLogger.WANDB:
                return [
                    WandbLogger(name=experiment_name, save_dir=work_dir / "wandb_logs"),
                    csv_logger,
                ]
            case SupportedLogger.TENSORBOARD:
                return [
                    TensorBoardLogger(save_dir=work_dir / "tb_logs"),
                    csv_logger,
                ]
            case _:
                return [csv_logger]

    def train(
        self,
        *,
        # BASIC PARAMS
        train_data: Any | None = None,
        train_data_target: Any | None = None,
        val_data: Any | None = None,
        val_data_target: Any | None = None,
        # val_percentage: float | None = None, # TODO: hidden till re-implemented
        # val_minimum_split: int = 5,
        # ADVANCED PARAMS
        filtering_mask: Any | None = None,
        read_source_func: Callable | None = None,
        read_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
    ) -> None:
        # TODO: init datamodule
        # TODO: remember to pass self.checkpoint_path to Trainer.fit
        # ^ this will load optimizer and lr_schedular state dicts
        raise NotImplementedError("Training is not implemented yet.")

    def predict(
        self,
        # BASIC PARAMS
        pred_data: Any | None = None,
        batch_size: int = 1,
        tile_size: tuple[int, ...] | None = None,
        tile_overlap: tuple[int, ...] | None = (48, 48),
        axes: str | None = None,
        data_type: Literal["array", "tiff", "custom"] | None = None,
        # ADVANCED PARAMS
        # tta_transforms: bool = False, # TODO: hidden till implemented
        num_workers: int | None = None,
        read_source_func: Callable | None = None,
        read_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
    ) -> None:
        raise NotImplementedError("Predicting is not implemented yet.")

    def predict_to_disk(
        self,
        # BASIC PARAMS
        pred_data: Any | None = None,
        pred_data_target: Any | None = None,
        prediction_dir: Path | str = "predictions",
        batch_size: int = 1,
        tile_size: tuple[int, ...] | None = None,
        tile_overlap: tuple[int, ...] | None = (48, 48),
        axes: str | None = None,
        data_type: Literal["array", "tiff", "custom"] | None = None,
        # ADVANCED PARAMS
        num_workers: int | None = None,
        read_source_func: Callable | None = None,
        read_kwargs: dict[str, Any] | None = None,
        extension_filter: str = "",
        # WRITE OPTIONS
        write_type: Literal["tiff", "zarr", "custom"] = "tiff",
        write_extension: str | None = None,
        write_func: WriteFunc | None = None,
        write_func_kwargs: dict[str, Any] | None = None,
    ) -> None:
        raise NotImplementedError("Predicting to disk is not implemented yet.")

    def export_to_bmz(
        self,
        path_to_archive: Path | str,
        friendly_model_name: str,
        input_array: NDArray,
        authors: list[dict],
        general_description: str,
        data_description: str,
        covers: list[Path | str] | None = None,
        channel_names: list[str] | None = None,
        model_version: str = "0.1.0",
    ) -> None:
        """Export the model to the BioImage Model Zoo format.

        This method packages the current weights into a zip file that can be uploaded
        to the BioImage Model Zoo. The archive consists of the model weights, the model
        specifications and various files (inputs, outputs, README, env.yaml etc.).

        `path_to_archive` should point to a file with a ".zip" extension.

        `friendly_model_name` is the name used for the model in the BMZ specs
        and website, it should consist of letters, numbers, dashes, underscores and
        parentheses only.

        Input array must be of the same dimensions as the axes recorded in the
        configuration of the `CAREamist`.

        Parameters
        ----------
        path_to_archive : pathlib.Path or str
            Path in which to save the model, including file name, which should end with
            ".zip".
        friendly_model_name : str
            Name of the model as used in the BMZ specs, it should consist of letters,
            numbers, dashes, underscores and parentheses only.
        input_array : NDArray
            Input array used to validate the model and as example.
        authors : list of dict
            List of authors of the model.
        general_description : str
            General description of the model used in the BMZ metadata.
        data_description : str
            Description of the data the model was trained on.
        covers : list of pathlib.Path or str, default=None
            Paths to the cover images.
        channel_names : list of str, default=None
            Channel names.
        model_version : str, default="0.1.0"
            Version of the model.
        """
        output_patch = self.predict(
            pred_data=input_array,
            data_type=SupportedData.ARRAY.value,
        )
        output = np.concatenate(output_patch, axis=0)
        input_array = reshape_array(input_array, self.config.data_config.axes)

        export_to_bmz(
            model=self.model,
            config=self.config,
            path_to_archive=path_to_archive,
            model_name=friendly_model_name,
            general_description=general_description,
            data_description=data_description,
            authors=authors,
            input_array=input_array,
            output_array=output,
            covers=covers,
            channel_names=channel_names,
            model_version=model_version,
        )

    def get_losses(self) -> dict[str, list]:
        """Return data that can be used to plot train and validation loss curves.

        Returns
        -------
        dict of str: list
            Dictionary containing losses for each epoch.
        """
        return read_csv_logger(self.config.experiment_name, self.work_dir / "csv_logs")

    def stop_training(self) -> None:
        """Stop the training loop."""
        self.trainer.should_stop = True
        self.trainer.limit_val_batches = 0 # skip validation



from pathlib import Path

import numpy as np
import pytest
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import (
    ModelCheckpoint,
)

from careamics.config import Configuration
from careamics.config.configuration_factories import create_pn2v_configuration
from careamics.lightning import (
    create_careamics_module,
    create_predict_datamodule,
    create_train_datamodule,
)
from careamics.lightning.lightning_module import FCNModule
from careamics.prediction_utils import convert_outputs, convert_outputs_pn2v

pytestmark = pytest.mark.mps_gh_fail


def test_smoke_n2v_2d_array(tmp_path, minimum_n2v_configuration):
    """Test a full run of N2V training with the lightning API."""
    rng = np.random.default_rng(42)

    # training data
    train_array = rng.integers(0, 255, (32, 32)).astype(np.float32)
    val_array = rng.integers(0, 255, (32, 32)).astype(np.float32)

    cfg = Configuration(**minimum_n2v_configuration)

    # create lightning module
    model = create_careamics_module(
        algorithm=cfg.algorithm_config.algorithm,
        loss=cfg.algorithm_config.loss,
        architecture=cfg.algorithm_config.model.architecture,
    )

    # create data module
    data = create_train_datamodule(
        train_data=train_array,
        val_data=val_array,
        data_type=cfg.data_config.data_type,
        patch_size=cfg.data_config.patch_size,
        axes=cfg.data_config.axes,
        batch_size=cfg.data_config.batch_size,
    )

    # create trainer
    trainer = Trainer(
        max_epochs=1,
        default_root_dir=tmp_path,
        callbacks=[
            ModelCheckpoint(
                dirpath=tmp_path / "checkpoints",
                filename="test_lightning_api",
            )
        ],
    )

    # train
    trainer.fit(model, datamodule=data)
    assert Path(tmp_path / "checkpoints" / "test_lightning_api.ckpt").exists()

    # predict
    means, stds = data.get_data_statistics()
    predict_data = create_predict_datamodule(
        pred_data=val_array,
        data_type=cfg.data_config.data_type,
        axes=cfg.data_config.axes,
        image_means=means,
        image_stds=stds,
    )

    # predict
    predicted = trainer.predict(model, datamodule=predict_data)
    assert predicted[0].squeeze().shape == val_array.shape


def test_smoke_n2v_2d_tiling(tmp_path, minimum_n2v_configuration):
    """Test a full run of N2V training with the lightning API and tiled prediction."""
    # training data
    rng = np.random.default_rng(42)
    train_array = rng.integers(0, 255, (32, 32)).astype(np.float32)
    val_array = rng.integers(0, 255, (32, 32)).astype(np.float32)

    cfg = Configuration(**minimum_n2v_configuration)

    # create lightning module
    model = create_careamics_module(
        algorithm=cfg.algorithm_config.algorithm,
        loss=cfg.algorithm_config.loss,
        architecture=cfg.algorithm_config.model.architecture,
    )

    # create data module
    data = create_train_datamodule(
        train_data=train_array,
        val_data=val_array,
        data_type=cfg.data_config.data_type,
        patch_size=cfg.data_config.patch_size,
        axes=cfg.data_config.axes,
        batch_size=cfg.data_config.batch_size,
    )

    # create trainer
    trainer = Trainer(
        max_epochs=1,
        default_root_dir=tmp_path,
        callbacks=[
            ModelCheckpoint(
                dirpath=tmp_path / "checkpoints",
                filename="test_lightning_api",
            )
        ],
    )

    # train
    trainer.fit(model, datamodule=data)
    assert Path(tmp_path / "checkpoints" / "test_lightning_api.ckpt").exists()

    # predict
    means, stds = data.get_data_statistics()
    predict_data = create_predict_datamodule(
        pred_data=val_array,
        data_type=cfg.data_config.data_type,
        axes=cfg.data_config.axes,
        image_means=means,
        image_stds=stds,
        tile_size=(8, 8),
        tile_overlap=(2, 2),
    )

    # predict
    predicted = trainer.predict(model, datamodule=predict_data)
    predicted_stitched = convert_outputs(predicted, tiled=True)
    assert predicted_stitched[0].shape[-2:] == val_array.shape


def test_smoke_pn2v_2d_array(tmp_path, create_dummy_noise_model):
    """Test a full run of PN2V training with the lightning API."""
    # Create dummy noise model file
    np.savez(tmp_path / "dummy_noise_model.npz", **create_dummy_noise_model)

    rng = np.random.default_rng(42)

    # training data
    train_array = rng.integers(0, 255, (128, 128)).astype(np.float32)
    val_array = rng.integers(0, 255, (128, 128)).astype(np.float32)

    # create PN2V configuration
    cfg = create_pn2v_configuration(
        experiment_name="test_pn2v",
        data_type="array",
        axes="YX",
        patch_size=[64, 64],
        batch_size=2,
        nm_path=str(tmp_path / "dummy_noise_model.npz"),
        num_epochs=1,
    )

    # create lightning module
    model = FCNModule(cfg.algorithm_config)

    # create data module
    data = create_train_datamodule(
        train_data=train_array,
        val_data=val_array,
        data_type=cfg.data_config.data_type,
        patch_size=cfg.data_config.patch_size,
        axes=cfg.data_config.axes,
        batch_size=cfg.data_config.batch_size,
    )

    # create trainer
    trainer = Trainer(
        max_epochs=1,
        default_root_dir=tmp_path,
        callbacks=[
            ModelCheckpoint(
                dirpath=tmp_path / "checkpoints",
                filename="test_pn2v_lightning_api",
            )
        ],
    )

    # train
    trainer.fit(model, datamodule=data)
    assert Path(tmp_path / "checkpoints" / "test_pn2v_lightning_api.ckpt").exists()

    # predict
    means, stds = data.get_data_statistics()
    predict_data = create_predict_datamodule(
        pred_data=val_array,
        data_type=cfg.data_config.data_type,
        axes=cfg.data_config.axes,
        image_means=means,
        image_stds=stds,
    )

    # predict
    predicted = trainer.predict(model, datamodule=predict_data)
    predicted_avg, mse = convert_outputs_pn2v(predicted, tiled=False)
    assert predicted_avg[0].squeeze().shape == val_array.shape
    assert mse[0].squeeze().shape == val_array.shape



"""Training and lightning related Pydantic configurations."""

__all__ = [
    "CheckpointConfig",
    "EarlyStoppingConfig",
    "LrSchedulerConfig",
    "OptimizerConfig",
    "TrainerConfig",
    "TrainingConfig",
]


from .callbacks import CheckpointConfig, EarlyStoppingConfig
from .optimizer_configs import LrSchedulerConfig, OptimizerConfig
from .training_config import TrainingConfig



"""Patching strategies supported by Careamics."""

from careamics.utils import BaseEnum


class SupportedPatchingStrategy(str, BaseEnum):
    """Patching strategies supported by Careamics."""

    FIXED_RANDOM = "fixed_random"
    """Fixed random patching strategy, used during training."""

    RANDOM = "random"
    """Random patching strategy, used during training."""

    STRATIFIED = "stratified"

    # SEQUENTIAL = "sequential"
    # """Sequential patching strategy, used during training."""

    TILED = "tiled"
    """Tiled patching strategy, used during prediction."""

    WHOLE = "whole"
    """Whole image patching strategy, used during prediction."""



from contextlib import nullcontext
from pathlib import Path
from typing import Literal

import numpy as np
import pytest
from numpy.typing import NDArray

from careamics.dataset_ng.image_stack.czi_image_stack import CziImageStack
from careamics.dataset_ng.image_stack.image_utils import channel_slice

# skip if fail imports
pylib = pytest.importorskip("pylibCZIrw")

from pylibCZIrw import czi as pyczi  # noqa: E402

T_EXPR = "does not contain a T axis"
Z_EXPR = "does not contain a Z axis"


def create_test_czi(file_path: Path, data: NDArray | list[NDArray]):
    if not isinstance(data, list):
        data = [data]

    with pyczi.create_czi(str(file_path)) as czi:
        xoffs = 0
        for scene_idx, scene_data in enumerate(data):
            for t in range(scene_data.shape[0]):
                for c in range(scene_data.shape[1]):
                    for z in range(scene_data.shape[2]):
                        czi.write(
                            scene_data[t, c, z],
                            plane={"C": c, "T": t, "Z": z},
                            location=(xoffs, 0),
                            scene=scene_idx,
                        )
            xoffs += scene_data.shape[-1] + 20


@pytest.mark.czi
class TestCziImageStack:

    @pytest.mark.parametrize(
        "orig_shape, depth_axis, expected_axes, expected_shape, sample_idx, "
        "expect_raise",
        [
            # 2-D Images
            ((1, 1, 1, 32, 48), "none", "SCYX", [1, 1, 32, 48], 0, nullcontext()),
            (
                (1, 1, 1, 32, 48),
                "T",
                "",
                [],
                0,
                pytest.raises(RuntimeError, match=T_EXPR),
            ),
            (
                (1, 1, 1, 32, 48),
                "Z",
                "",
                [],
                0,
                pytest.raises(RuntimeError, match=Z_EXPR),
            ),
            # 3-D Volumes
            ((1, 1, 16, 32, 48), "none", "SCYX", [16, 1, 32, 48], 9, nullcontext()),
            (
                (1, 1, 16, 32, 48),
                "T",
                "",
                [],
                9,
                pytest.raises(RuntimeError, match=T_EXPR),
            ),
            ((1, 1, 16, 32, 48), "Z", "SCZYX", [1, 1, 16, 32, 48], 0, nullcontext()),
            # 2-D Time-Series
            ((8, 1, 1, 32, 48), "none", "SCYX", [8, 1, 32, 48], 3, nullcontext()),
            (
                (8, 1, 1, 32, 48),
                "Z",
                "",
                [],
                3,
                pytest.raises(RuntimeError, match=Z_EXPR),
            ),
            ((8, 1, 1, 32, 48), "T", "SCTYX", [1, 1, 8, 32, 48], 0, nullcontext()),
            # 3-D Time-Series
            (
                (8, 1, 16, 32, 48),
                "none",
                "SCYX",
                [8 * 16, 1, 32, 48],
                35,
                nullcontext(),
            ),
            ((8, 1, 16, 32, 48), "Z", "SCZYX", [8, 1, 16, 32, 48], 7, nullcontext()),
            ((8, 1, 16, 32, 48), "T", "SCTYX", [16, 1, 8, 32, 48], 12, nullcontext()),
            # Multiple channels
            ((8, 3, 16, 32, 48), "Z", "SCZYX", [8, 3, 16, 32, 48], 2, nullcontext()),
        ],
    )
    def test_extract_patch(
        self,
        tmp_path: Path,
        orig_shape: tuple[int, ...],
        depth_axis: Literal["none", "Z", "T"],
        expected_axes: str,
        expected_shape: list[int],
        sample_idx: int,
        expect_raise,
    ):
        # reference data to compare against
        data = np.random.randn(*orig_shape).astype(np.float32)

        # save data as a czi file to ininitialise image stack with
        file_path = tmp_path / "test_czi.czi"
        create_test_czi(file_path=file_path, data=data)

        # initialise CziImageStack
        with expect_raise:
            image_stack = CziImageStack(data_path=file_path, depth_axis=depth_axis)

        # stop here if expecting an exception
        if expected_axes == "":
            return

        # check axes and shape
        assert image_stack.axes == expected_axes
        assert image_stack.data_shape == expected_shape

        # test extracted patch matches patch from reference data
        if len(expected_axes) < 5:
            coords = (11, 4)
            patch_size = (16, 9)

            extracted_patch = image_stack.extract_patch(
                sample_idx=sample_idx,
                channels=None,
                coords=coords,
                patch_size=patch_size,
            )

            data_ref = np.moveaxis(data, 2, 1)  # (T, Z, C, Y, X)
            data_ref = data_ref.reshape(-1, *data_ref.shape[-3:])
            patch_ref = data_ref[
                sample_idx,
                :,
                coords[0] : coords[0] + patch_size[0],
                coords[1] : coords[1] + patch_size[1],
            ]
        else:
            coords = (2, 11, 4)
            patch_size = (4, 16, 9)

            extracted_patch = image_stack.extract_patch(
                sample_idx=sample_idx,
                channels=None,
                coords=coords,
                patch_size=patch_size,
            )

            if "T" in expected_axes and expected_axes.index("T") == 2:
                data_ref = data.swapaxes(0, 2)
            else:
                data_ref = data
            patch_ref = data_ref[
                sample_idx,
                ...,
                coords[0] : coords[0] + patch_size[0],
                coords[1] : coords[1] + patch_size[1],
                coords[2] : coords[2] + patch_size[2],
            ]
        np.testing.assert_array_equal(extracted_patch, patch_ref)

    def test_multiple_scenes(
        self,
        tmp_path: Path,
    ):
        original_shapes = [
            [4, 1, 8, 16, 32],
            [4, 1, 8, 24, 18],
            [4, 1, 8, 45, 32],
        ]

        # reference data to compare against
        data_ref = [
            np.random.randn(*shape).astype(np.float32) for shape in original_shapes
        ]

        # save data as a czi file to ininitialise image stack with
        file_path = tmp_path / "test_czi.czi"
        create_test_czi(file_path=file_path, data=data_ref)

        # Test reading scene metadata
        scene_rectangles = CziImageStack.get_bounding_rectangles(file_path)
        assert len(scene_rectangles) == len(original_shapes)

        scene_rectangle_sizes = [(rect.h, rect.w) for rect in scene_rectangles.values()]
        expected_rectangle_sizes = [(shape[-2], shape[-1]) for shape in original_shapes]
        assert scene_rectangle_sizes == expected_rectangle_sizes

        # Test reading from individual scenes
        for scene_idx, expected_shape in enumerate(original_shapes):
            image_stack = CziImageStack(
                data_path=file_path, scene=scene_idx, depth_axis="Z"
            )
            assert image_stack.data_shape == expected_shape

            t = expected_shape[0] - scene_idx - 1
            coords = (2, 9, 4)
            patch_size = (4, 7, 13)

            extracted_patch = image_stack.extract_patch(
                sample_idx=t, channels=None, coords=coords, patch_size=patch_size
            )
            patch_ref = data_ref[scene_idx][
                t,
                :,
                coords[0] : coords[0] + patch_size[0],
                coords[1] : coords[1] + patch_size[1],
                coords[2] : coords[2] + patch_size[2],
            ]
            np.testing.assert_array_equal(extracted_patch, patch_ref)


@pytest.mark.czi
@pytest.mark.parametrize("axis", ["Z", "T"])
def test_z_padding(tmp_path: Path, axis):
    """Test that Z padding is applied to CziImageStack when requesting data outside
    the Z range."""
    if "Z" == axis:
        shape = (1, 2, 4, 4, 4)  # (T, C, Z, Y, X)
    else:
        shape = (4, 2, 1, 4, 4)  # (Z, C, T, Y, X)

    data = np.random.randn(*shape).astype(np.float32)

    file_path = tmp_path / "test_czi.czi"
    create_test_czi(file_path=file_path, data=data)

    # initialise CziImageStack
    image_stack = CziImageStack(data_path=file_path, depth_axis=axis)

    # extract patches
    sample_idx = 0
    channel_idx = 1
    patch_size = (4, 4, 4)
    z_length = shape[2] if "Z" == axis else shape[0]
    coordinates = [(-2, 0, 0), (2, 0, 0)]
    for coord in coordinates:
        patch = image_stack.extract_patch(
            sample_idx=sample_idx,
            channels=[channel_idx],
            coords=coord,
            patch_size=patch_size,
        )

        # build expected array
        z_start = max(coord[0], 0)
        z_end = min(coord[0] + patch_size[0], z_length)
        z_start_pad = z_start - coord[0]
        z_end_pad = z_end - coord[0]

        expected_patch = np.zeros(patch_size, dtype=np.float32)

        if "Z" == axis:
            expected_patch[z_start_pad:z_end_pad, :, :] = data[
                sample_idx, channel_idx, z_start:z_end, :, :
            ]
        else:  # T
            expected_patch[z_start_pad:z_end_pad, :, :] = data[
                z_start:z_end, channel_idx, sample_idx, :, :
            ]

        np.testing.assert_array_equal(patch[0], expected_patch)


@pytest.mark.czi
class TestCziImageStackChannels:

    @pytest.mark.parametrize(
        "shape, channels",
        [
            ((1, 1, 1, 32, 32), None),
            ((1, 1, 1, 32, 32), [0]),
            ((1, 3, 1, 32, 32), None),
            ((1, 3, 1, 32, 32), [0, 2]),
            ((1, 3, 1, 32, 32), [2]),
        ],
    )
    def test_extract_channels(
        self,
        tmp_path: Path,
        shape: tuple[int, ...],
        channels: list[int] | None,
    ):
        # reference data to compare against
        data = np.random.randn(*shape).astype(np.float32)

        # save data as a czi file to initialize image stack with
        file_path = tmp_path / "test_czi.czi"
        create_test_czi(file_path=file_path, data=data)

        # initialise CziImageStack
        image_stack = CziImageStack(data_path=file_path)

        # extract patch
        patch = image_stack.extract_patch(
            sample_idx=0,
            channels=channels,
            coords=(0, 0),
            patch_size=(8, 8),
        )
        assert len(patch.shape) == 3  # no Z
        assert (
            patch.shape[0] == len(channels) if channels is not None else data.shape[0]
        )

        expected_patch = data[
            0,
            channel_slice(channels),
            0,
            0 : 0 + 8,
            0 : 0 + 8,
        ]
        np.testing.assert_array_equal(patch, expected_patch)

    @pytest.mark.parametrize(
        "shape, channels",
        [
            ((2, 3, 1, 64, 64), [0, 4]),
            ((2, 3, 1, 64, 64), [3]),
        ],
    )
    def test_extract_channel_error(
        self,
        tmp_path: Path,
        shape: tuple[int, ...],
        channels: int,
    ):
        # reference data to compare against
        data = np.random.randn(*shape).astype(np.float32)

        # save data as a czi file to ininitialise image stack with
        file_path = tmp_path / "test_czi.czi"
        create_test_czi(file_path=file_path, data=data)

        # initialise CziImageStack
        image_stack = CziImageStack(data_path=file_path)

        expected_msg = (
            f"Channel index {channels[-1]} is out of bounds for data with "
            f"{shape[1]} channels. Check the provided `channels` "
            f"parameter in the configuration for erroneous channel "
            f"indices."
        )

        with pytest.raises(ValueError, match=expected_msg):
            image_stack.extract_patch(
                sample_idx=0,
                channels=channels,
                coords=(0, 0),
                patch_size=(16, 16),
            )



"""Convenience functions to create NG data configurations."""

from collections.abc import Sequence
from typing import Any, Literal

from careamics.config.data import NGDataConfig
from careamics.config.transformations import (
    SPATIAL_TRANSFORMS_UNION,
    XYFlipConfig,
    XYRandomRotate90Config,
)


def list_spatial_augmentations(
    augmentations: list[SPATIAL_TRANSFORMS_UNION] | None = None,
) -> list[SPATIAL_TRANSFORMS_UNION]:
    """
    List the augmentations to apply.

    Parameters
    ----------
    augmentations : list of transforms, optional
        List of transforms to apply, either both or one of XYFlipConfig and
        XYRandomRotate90Config.

    Returns
    -------
    list of transforms
        List of transforms to apply.

    Raises
    ------
    ValueError
        If the transforms are not XYFlipConfig or XYRandomRotate90Config.
    ValueError
        If there are duplicate transforms.
    """
    if augmentations is None:
        transform_list: list[SPATIAL_TRANSFORMS_UNION] = [
            XYFlipConfig(),
            XYRandomRotate90Config(),
        ]
    else:
        # throw error if not all transforms are pydantic models
        if not all(
            isinstance(t, XYFlipConfig) or isinstance(t, XYRandomRotate90Config)
            for t in augmentations
        ):
            raise ValueError(
                "Accepted transforms are either XYFlipConfig or "
                "XYRandomRotate90Config."
            )

        # check that there is no duplication
        aug_types = [t.__class__ for t in augmentations]
        if len(set(aug_types)) != len(aug_types):
            raise ValueError("Duplicate transforms are not allowed.")

        transform_list = augmentations

    return transform_list


def create_ng_data_configuration(
    data_type: Literal["array", "tiff", "zarr", "czi", "custom"],
    axes: str,
    patch_size: Sequence[int],
    batch_size: int,
    augmentations: list[SPATIAL_TRANSFORMS_UNION] | None = None,
    normalization: dict | None = None,
    channels: Sequence[int] | None = None,
    in_memory: bool | None = None,
    num_workers: int = 0,
    train_dataloader_params: dict[str, Any] | None = None,
    val_dataloader_params: dict[str, Any] | None = None,
    pred_dataloader_params: dict[str, Any] | None = None,
    seed: int | None = None,
) -> NGDataConfig:
    """
    Create a training NGDatasetConfig.

    Note that `num_workers` is applied to all dataloaders unless explicitly overridden
    in the respective dataloader parameters.

    Parameters
    ----------
    data_type : {"array", "tiff", "zarr", "czi", "custom"}
        Type of the data.
    axes : str
        Axes of the data.
    patch_size : list of int
        Size of the patches along the spatial dimensions.
    batch_size : int
        Batch size.
    augmentations : list of transforms or None, default=None
        List of transforms to apply. If `None`, default augmentations are applied
        (flip in X and Y, rotations by 90 degrees in the XY plane).
    normalization : dict, default=None
        Normalization configuration dictionary. If None, defaults to mean_std
        normalization with automatically computed statistics.
    channels : Sequence of int, default=None
        List of channels to use. If `None`, all channels are used.
    in_memory : bool, default=None
        Whether to load all data into memory. This is only supported for 'array',
        'tiff' and 'custom' data types. If `None`, defaults to `True` for 'array',
        'tiff' and `custom`, and `False` for 'zarr' and 'czi' data types. Must be `True`
        for `array`.
    num_workers : int, default=0
        Number of workers for data loading.
    augmentations : list of transforms or None, default=None
        List of transforms to apply. If `None`, default augmentations are applied
        (flip in X and Y, rotations by 90 degrees in the XY plane).
    train_dataloader_params : dict
        Parameters for the training dataloader, see PyTorch notes, by default None.
    val_dataloader_params : dict
        Parameters for the validation dataloader, see PyTorch notes, by default None.
    pred_dataloader_params : dict
        Parameters for the test dataloader, see PyTorch notes, by default None.
    seed : int, default=None
        Random seed for reproducibility. If `None`, no seed is set.

    Returns
    -------
    NGDataConfig
        Next-Generation Data model with the specified parameters.
    """
    if augmentations is None:
        augmentations = list_spatial_augmentations()

    # data model
    data: dict[str, Any] = {
        "mode": "training",
        "data_type": data_type,
        "axes": axes,
        "batch_size": batch_size,
        "channels": channels,
        "transforms": augmentations,
        "seed": seed,
        "normalization": (
            normalization if normalization is not None else {"name": "mean_std"}
        ),
    }

    if in_memory is not None:
        data["in_memory"] = in_memory

    if train_dataloader_params is not None:
        # the presence of `shuffle` key in the dataloader parameters is enforced
        # by the NGDataConfig class
        if "shuffle" not in train_dataloader_params:
            train_dataloader_params["shuffle"] = True

        if "num_workers" not in train_dataloader_params:
            train_dataloader_params["num_workers"] = num_workers

        data["train_dataloader_params"] = train_dataloader_params
    else:
        data["train_dataloader_params"] = {"shuffle": True, "num_workers": num_workers}

    if val_dataloader_params is not None:
        if "num_workers" not in val_dataloader_params:
            val_dataloader_params["num_workers"] = num_workers

        data["val_dataloader_params"] = val_dataloader_params
    else:
        data["val_dataloader_params"] = {"shuffle": False, "num_workers": num_workers}

    if pred_dataloader_params is not None:
        if "num_workers" not in pred_dataloader_params:
            pred_dataloader_params["num_workers"] = num_workers

        data["pred_dataloader_params"] = pred_dataloader_params
    else:
        data["pred_dataloader_params"] = {"shuffle": False, "num_workers": num_workers}

    # add training patching
    data["patching"] = {
        "name": "random",
        "patch_size": patch_size,
    }

    return NGDataConfig(**data)



"""
A place for Datasets and Dataloaders.
"""

import logging
import math
from pathlib import Path
from typing import Any, Callable, Optional, Union

import numpy as np
from skimage.transform import resize

from .config import MicroSplitDataConfig
from .multich_dataset import MultiChDloader


class LCMultiChDloader(MultiChDloader):
    """Multi-channel dataset loader for LC-style datasets."""

    def __init__(
        self,
        data_config: MicroSplitDataConfig,
        datapath: Union[str, Path],
        load_data_fn: Optional[Callable] = None,
        val_fraction: float = 0.1,
        test_fraction: float = 0.1,
        allow_generation: bool = False,
    ):
        self._padding_kwargs = (
            data_config.padding_kwargs  # mode=padding_mode, constant_values=constant_value
        )
        self._uncorrelated_channel_probab = data_config.uncorrelated_channel_probab

        super().__init__(
            data_config,
            datapath,
            load_data_fn=load_data_fn,
            val_fraction=val_fraction,
            test_fraction=test_fraction,
        )

        if data_config.overlapping_padding_kwargs is not None:
            assert (
                self._padding_kwargs == data_config.overlapping_padding_kwargs
            ), "During evaluation, overlapping_padding_kwargs should be same as padding_args. \
                It should be so since we just use overlapping_padding_kwargs when it is not None"

        else:
            self._overlapping_padding_kwargs = data_config.padding_kwargs

        self.multiscale_lowres_count = data_config.multiscale_lowres_count
        assert self.multiscale_lowres_count is not None
        self._scaled_data = [self._data]
        self._scaled_noise_data = [self._noise_data]

        assert (
            isinstance(self.multiscale_lowres_count, int)
            and self.multiscale_lowres_count >= 1
        )
        assert isinstance(self._padding_kwargs, dict)
        assert "mode" in self._padding_kwargs

        for _ in range(1, self.multiscale_lowres_count):
            shape = self._scaled_data[-1].shape
            assert len(shape) == 4
            new_shape = (shape[0], shape[1] // 2, shape[2] // 2, shape[3])
            ds_data = resize(
                self._scaled_data[-1].astype(np.float32), new_shape
            ).astype(self._scaled_data[-1].dtype)
            # NOTE: These asserts are important. the resize method expects np.float32. otherwise, one gets weird results.
            assert (
                ds_data.max() / self._scaled_data[-1].max() < 5
            ), "Downsampled image should not have very different values"
            assert (
                ds_data.max() / self._scaled_data[-1].max() > 0.2
            ), "Downsampled image should not have very different values"

            self._scaled_data.append(ds_data)
            # do the same for noise
            if self._noise_data is not None:
                noise_data = resize(self._scaled_noise_data[-1], new_shape)
                self._scaled_noise_data.append(noise_data)

    def reduce_data(
        self, t_list=None, h_start=None, h_end=None, w_start=None, w_end=None
    ):
        assert t_list is not None
        assert h_start is None
        assert h_end is None
        assert w_start is None
        assert w_end is None

        self._data = self._data[t_list].copy()
        self._scaled_data = [
            self._scaled_data[i][t_list].copy() for i in range(len(self._scaled_data))
        ]

        if self._noise_data is not None:
            self._noise_data = self._noise_data[t_list].copy()
            self._scaled_noise_data = [
                self._scaled_noise_data[i][t_list].copy()
                for i in range(len(self._scaled_noise_data))
            ]

        self.N = len(t_list)
        # TODO where tf is self._img_sz defined?
        self.set_img_sz([self._img_sz, self._img_sz], self._grid_sz)
        print(
            f"[{self.__class__.__name__}] Data reduced. New data shape: {self._data.shape}"
        )

    def _init_msg(self):
        msg = super()._init_msg()
        msg += f" Pad:{self._padding_kwargs}"
        if self._uncorrelated_channels:
            msg += f" UncorrChProbab:{self._uncorrelated_channel_probab}"
        return msg

    def _load_scaled_img(
        self, scaled_index, index: Union[int, tuple[int, int]]
    ) -> tuple[np.ndarray, np.ndarray]:
        if isinstance(index, int):
            idx = index
        else:
            idx, _ = index

        # tidx = self.idx_manager.get_t(idx)
        patch_loc_list = self.idx_manager.get_patch_location_from_dataset_idx(idx)
        nidx = patch_loc_list[0]

        imgs = self._scaled_data[scaled_index][nidx]
        imgs = tuple([imgs[None, ..., i] for i in range(imgs.shape[-1])])
        if self._noise_data is not None:
            noisedata = self._scaled_noise_data[scaled_index][nidx]
            noise = tuple([noisedata[None, ..., i] for i in range(noisedata.shape[-1])])
            factor = np.sqrt(2) if self._input_is_sum else 1.0
            imgs = tuple([img + noise[0] * factor for img in imgs])
        return imgs

    def _crop_img(self, img: np.ndarray, patch_start_loc: tuple):
        """
        Here, h_start, w_start could be negative. That simply means we need to pick the content from 0. So,
        the cropped image will be smaller than self._img_sz * self._img_sz
        """
        max_len_vals = list(self.idx_manager.data_shape[1:-1])
        max_len_vals[-2:] = img.shape[-2:]
        return self._crop_img_with_padding(
            img, patch_start_loc, max_len_vals=max_len_vals
        )

    def _get_img(self, index: int):
        """
        Returns the primary patch along with low resolution patches centered on the primary patch.
        """
        # Noise_tuples is populated when there is synthetic noise in training
        # Should have similar type of noise with the noise model
        # Starting with microsplit, dump the noise, use it instead as an augmentation if nessesary
        img_tuples, noise_tuples = self._load_img(index)
        assert self._img_sz is not None
        h, w = img_tuples[0].shape[-2:]
        if self._enable_random_cropping:
            patch_start_loc = self._get_random_hw(h, w)
            if self._5Ddata:
                patch_start_loc = (
                    np.random.choice(img_tuples[0].shape[-3] - self._depth3D),
                ) + patch_start_loc
        else:
            patch_start_loc = self._get_deterministic_loc(index)

        # LC logic is located here, the function crops the image of the highest resolution
        cropped_img_tuples = [
            self._crop_flip_img(img, patch_start_loc, False, False)
            for img in img_tuples
        ]
        cropped_noise_tuples = [
            self._crop_flip_img(noise, patch_start_loc, False, False)
            for noise in noise_tuples
        ]
        patch_start_loc = list(patch_start_loc)
        h_start, w_start = patch_start_loc[-2], patch_start_loc[-1]
        h_center = h_start + self._img_sz // 2
        w_center = w_start + self._img_sz // 2
        allres_versions = {
            i: [cropped_img_tuples[i]] for i in range(len(cropped_img_tuples))
        }
        for scale_idx in range(1, self.multiscale_lowres_count):
            # Returning the image of the lower resolution
            scaled_img_tuples = self._load_scaled_img(scale_idx, index)

            h_center = h_center // 2
            w_center = w_center // 2

            h_start = h_center - self._img_sz // 2
            w_start = w_center - self._img_sz // 2
            patch_start_loc[-2:] = [h_start, w_start]
            scaled_cropped_img_tuples = [
                self._crop_flip_img(img, patch_start_loc, False, False)
                for img in scaled_img_tuples
            ]
            for ch_idx in range(len(img_tuples)):
                allres_versions[ch_idx].append(scaled_cropped_img_tuples[ch_idx])

        output_img_tuples = tuple(
            [
                np.concatenate(allres_versions[ch_idx])
                for ch_idx in range(len(img_tuples))
            ]
        )
        return output_img_tuples, cropped_noise_tuples

    def __getitem__(self, index: Union[int, tuple[int, int]]):
        img_tuples, noise_tuples = self._get_img(index)
        if self._uncorrelated_channels:
            assert (
                self._input_idx is None
            ), "Uncorrelated channels is not implemented when there is a separate input channel."
            if np.random.rand() < self._uncorrelated_channel_probab:
                img_tuples_new = [None] * len(img_tuples)
                img_tuples_new[0] = img_tuples[0]
                for i in range(1, len(img_tuples)):
                    new_index = np.random.randint(len(self))
                    img_tuples_tmp, _ = self._get_img(new_index)
                    img_tuples_new[i] = img_tuples_tmp[i]
                img_tuples = img_tuples_new

        if self._is_train:
            if self._empty_patch_replacement_enabled:
                if np.random.rand() < self._empty_patch_replacement_probab:
                    img_tuples = self.replace_with_empty_patch(img_tuples)

        if self._enable_rotation:
            img_tuples, noise_tuples = self._rotate(img_tuples, noise_tuples)

        # add noise to input, if noise is present combine it with the image
        # factor is for the compute input not to have too much noise because the average of two gaussians
        if len(noise_tuples) > 0:
            factor = np.sqrt(2) if self._input_is_sum else 1.0
            input_tuples = []
            for x in img_tuples:
                x = (
                    x.copy()
                )  # to avoid changing the original image since it is later used for target
                # NOTE: other LC levels already have noise added. So, we just need to add noise to the highest resolution.
                x[0] = x[0] + noise_tuples[0] * factor
                input_tuples.append(x)
        else:
            input_tuples = img_tuples

        # Compute the input by sum / average the channels
        # Alpha is an amount of weight which is applied to the channels when combining them
        # How to sample alpha is still under research
        inp, alpha = self._compute_input(input_tuples)
        target_tuples = [img[:1] for img in img_tuples]
        # add noise to target.
        if len(noise_tuples) >= 1:
            target_tuples = [
                x + noise for x, noise in zip(target_tuples, noise_tuples[1:])
            ]

        target = self._compute_target(target_tuples, alpha)

        norm_target = self.normalize_target(target)

        output = [inp, norm_target]

        if self._return_alpha:
            output.append(alpha)

        if isinstance(index, int):
            return tuple(output)

        _, grid_size = index
        output.append(grid_size)
        return tuple(output)



"""Sequential patching Pydantic model."""

from collections.abc import Sequence

from pydantic import Field, ValidationInfo, field_validator

from ._patched_config import _PatchedConfig


class _OverlappingPatchedConfig(_PatchedConfig):
    """Overlapping patching Pydantic model.

    This model is only used for inheritance and validation purposes.

    Attributes
    ----------
    patch_size : list of int
        The size of the patch in each spatial dimension, each patch size must be a power
        of 2 and larger than 8.
    overlaps : sequence of int, optional
        The overlaps between patches in each spatial dimension. If `None`, no overlap is
        applied. The overlaps must be smaller than the patch size in each spatial
        dimension, and the number of dimensions be either 2 or 3.
    """

    overlaps: Sequence[int] | None = Field(
        default=None,
        min_length=2,
        max_length=3,
    )
    """The overlaps between patches in each spatial dimension. If `None`, no overlap is
    applied. The overlaps must be smaller than the patch size in each spatial dimension,
    and the number of dimensions be either 2 or 3.
    """

    @field_validator("overlaps")
    @classmethod
    def overlap_smaller_than_patch_size(
        cls, overlaps: Sequence[int] | None, values: ValidationInfo
    ) -> Sequence[int] | None:
        """
        Validate overlap.

        Overlaps must be smaller than the patch size in each spatial dimension.

        Parameters
        ----------
        overlaps : Sequence of int
            Overlap in each dimension.
        values : ValidationInfo
            Dictionary of values.

        Returns
        -------
        Sequence of int
            Validated overlap.
        """
        if overlaps is None:
            return None

        if "patch_size" not in values.data:
            raise ValueError(
                "Cannot validate overlaps because of undefined patch size. There may  "
                "be validation errors in `patch_size`."
            )

        patch_size = values.data["patch_size"]

        if len(overlaps) != len(patch_size):
            raise ValueError(
                f"Overlaps must have the same number of dimensions as the patch size. "
                f"Got {len(overlaps)} dimensions for overlaps and {len(patch_size)} "
                f"dimensions for patch size."
            )

        if any(o >= p for o, p in zip(overlaps, patch_size, strict=False)):
            raise ValueError(
                f"Overlap must be smaller than the patch size, got {overlaps} versus "
                f"{patch_size}."
            )

        return overlaps

    @field_validator("overlaps")
    @classmethod
    def overlap_even(cls, overlaps: Sequence[int] | None) -> Sequence[int] | None:
        """
        Validate overlaps.

        Overlap must be even.

        Parameters
        ----------
        overlaps : Sequence of int
            Overlaps.

        Returns
        -------
        Sequence of int
            Validated overlap.
        """
        if overlaps is None:
            return None

        if any(o % 2 != 0 for o in overlaps):
            raise ValueError(f"Overlaps must be even, got {overlaps}.")

        return overlaps



"""VAE-based algorithm Pydantic model."""

from __future__ import annotations

from pprint import pformat
from typing import Literal, Self

from pydantic import BaseModel, ConfigDict, model_validator

from careamics.config.architectures import LVAEConfig
from careamics.config.lightning.optimizer_configs import (
    LrSchedulerConfig,
    OptimizerConfig,
)
from careamics.config.losses.loss_config import LVAELossConfig
from careamics.config.noise_model.likelihood_config import (
    GaussianLikelihoodConfig,
    NMLikelihoodConfig,
)
from careamics.config.noise_model.noise_model_config import MultiChannelNMConfig
from careamics.config.support import SupportedAlgorithm, SupportedLoss


class VAEBasedAlgorithm(BaseModel):
    """VAE-based algorithm configuration.

    # TODO

    Examples
    --------
    # TODO add once finalized
    """

    # Pydantic class configuration
    model_config = ConfigDict(
        protected_namespaces=(),  # allows to use model_* as a field name
        validate_assignment=True,
        extra="allow",
    )

    # Mandatory fields
    # defined in SupportedAlgorithm
    # TODO: Use supported Enum classes for typing?
    #   - values can still be passed as strings and they will be cast to Enum
    algorithm: Literal["hdn", "microsplit"]

    # NOTE: these are all configs (pydantic models)
    loss: LVAELossConfig
    model: LVAEConfig
    noise_model: MultiChannelNMConfig | None = None
    noise_model_likelihood: NMLikelihoodConfig | None = None
    gaussian_likelihood: GaussianLikelihoodConfig | None = None  # TODO change to str

    mmse_count: int = 1
    is_supervised: bool = False

    # Optional fields
    optimizer: OptimizerConfig = OptimizerConfig()
    """Optimizer to use, defined in SupportedOptimizer."""

    lr_scheduler: LrSchedulerConfig = LrSchedulerConfig()

    @model_validator(mode="after")
    def algorithm_cross_validation(self: Self) -> Self:
        """Validate the algorithm model based on `algorithm`.

        Returns
        -------
        Self
            The validated model.
        """
        # hdn
        # TODO move to designated configurations
        if self.algorithm == SupportedAlgorithm.HDN:
            if self.loss.loss_type != SupportedLoss.HDN:
                raise ValueError(
                    f"Algorithm {self.algorithm} only supports loss `hdn`."
                )
            if self.model.multiscale_count > 1:
                raise ValueError("Algorithm `hdn` does not support multiscale models.")
        # musplit
        if self.algorithm == SupportedAlgorithm.MICROSPLIT:
            if self.loss.loss_type not in [
                SupportedLoss.MUSPLIT,
                SupportedLoss.DENOISPLIT,
                SupportedLoss.DENOISPLIT_MUSPLIT,
            ]:  # TODO Update losses configs, make loss just microsplit
                raise ValueError(
                    f"Algorithm {self.algorithm} only supports loss `microsplit`."
                )  # TODO Update losses configs

            if (
                self.loss.loss_type == SupportedLoss.DENOISPLIT
                and self.model.predict_logvar is not None
            ):
                raise ValueError(
                    "Algorithm `denoisplit` with loss `denoisplit` only supports "
                    "`predict_logvar` as `None`."
                )
            if (
                self.loss.loss_type == SupportedLoss.DENOISPLIT
                and self.noise_model is None
            ):
                raise ValueError("Algorithm `denoisplit` requires a noise model.")
        # TODO: what if algorithm is not musplit or denoisplit
        return self

    @model_validator(mode="after")
    def output_channels_validation(self: Self) -> Self:
        """Validate the consistency between number of out channels and noise models.

        Returns
        -------
        Self
            The validated model.
        """
        if self.noise_model is not None:
            assert self.model.output_channels == len(self.noise_model.noise_models), (
                f"Number of output channels ({self.model.output_channels}) must match "
                f"the number of noise models ({len(self.noise_model.noise_models)})."
            )

        if self.algorithm == SupportedAlgorithm.HDN:
            assert self.model.output_channels == 1, (
                f"Number of output channels ({self.model.output_channels}) must be 1 "
                "for algorithm `hdn`."
            )
        return self

    @model_validator(mode="after")
    def predict_logvar_validation(self: Self) -> Self:
        """Validate the consistency of `predict_logvar` throughout the model.

        Returns
        -------
        Self
            The validated model.
        """
        if self.gaussian_likelihood is not None:
            assert (
                self.model.predict_logvar == self.gaussian_likelihood.predict_logvar
            ), (
                f"Model `predict_logvar` ({self.model.predict_logvar}) must match "
                "Gaussian likelihood model `predict_logvar` "
                f"({self.gaussian_likelihood.predict_logvar}).",
            )
        # if self.algorithm == SupportedAlgorithm.HDN:
        #     assert (
        #         self.model.predict_logvar is None
        #     ), "Model `predict_logvar` must be `None` for algorithm `hdn`."
        #     if self.gaussian_likelihood is not None:
        #         assert self.gaussian_likelihood.predict_logvar is None, (
        #             "Gaussian likelihood model `predict_logvar` must be `None` "
        #             "for algorithm `hdn`."
        #         )
        # TODO check this
        return self

    def __str__(self) -> str:
        """Pretty string representing the configuration.

        Returns
        -------
        str
            Pretty string.
        """
        return pformat(self.model_dump())

    @classmethod
    def get_compatible_algorithms(cls) -> list[str]:
        """Get the list of compatible algorithms.

        Returns
        -------
        list of str
            List of compatible algorithms.
        """
        return ["hdn", "microsplit"]



"""Test prediction conversion and callback."""

import numpy as np
import pytest
import tifffile
import zarr
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import ModelCheckpoint

from careamics.config import N2VAlgorithm, create_ng_data_configuration
from careamics.lightning.dataset_ng.callbacks.prediction_writer import (
    PredictionWriterCallback,
)
from careamics.lightning.dataset_ng.data_module import CareamicsDataModule
from careamics.lightning.dataset_ng.lightning_modules import N2VModule
from careamics.lightning.dataset_ng.prediction import convert_prediction

# TODO add zarr writing from tiff and arrays when it will be available
# TODO add untiled zarr writing when it will be implemented


@pytest.mark.mps_gh_fail
@pytest.mark.parametrize(
    "shape, axes, channels, tiled",
    [
        ((32, 32), "YX", None, True),
        ((32, 32), "YX", None, False),
        ((3, 32, 32), "CYX", None, True),
        ((3, 32, 32), "CYX", None, False),
        ((3, 32, 32), "CYX", [1], True),
        ((3, 32, 32), "CYX", [1], False),
        ((3, 32, 32), "CYX", [0, 2], True),
        ((3, 32, 32), "CYX", [0, 2], False),
        ((16, 32, 32), "ZYX", None, True),
        ((16, 32, 32), "ZYX", None, False),
        ((5, 16, 32, 32), "SZYX", None, True),
        ((5, 16, 32, 32), "SZYX", None, False),
        ((3, 16, 32, 32), "CZYX", None, True),
        ((3, 16, 32, 32), "CZYX", None, False),
        ((3, 16, 32, 32), "CZYX", [1], True),
        ((3, 16, 32, 32), "CZYX", [1], False),
        ((3, 16, 32, 32), "CZYX", [0, 2], True),
        ((3, 16, 32, 32), "CZYX", [0, 2], False),
        ((5, 3, 16, 32, 32), "SCZYX", None, True),
        ((5, 3, 16, 32, 32), "SCZYX", None, False),
        ((5, 3, 16, 32, 32), "SCZYX", [1], True),
        ((5, 3, 16, 32, 32), "SCZYX", [1], False),
        ((5, 3, 16, 32, 32), "SCZYX", [0, 2], True),
        ((5, 3, 16, 32, 32), "SCZYX", [0, 2], False),
    ],
)
def test_smoke_n2v_tiff(tmp_path, shape, axes, channels, tiled):
    rng = np.random.default_rng(42)

    # training data
    train_array = rng.integers(0, 255, shape).astype(np.float32)
    val_array = rng.integers(0, 255, shape).astype(np.float32)

    # write train array to tiff
    train_dir = tmp_path / "train"
    train_dir.mkdir()
    file_name = "image.tiff"
    train_file = train_dir / file_name
    tifffile.imwrite(train_file, train_array)

    if "C" in axes:
        len_channels = shape[axes.index("C")]
    else:
        len_channels = 1

    algorithm_config = N2VAlgorithm(
        model={
            "architecture": "UNet",
            "conv_dims": 2 if "Z" not in axes else 3,
            "in_channels": len(channels) if channels is not None else len_channels,
            "num_classes": len(channels) if channels is not None else len_channels,
        }
    )
    # create NGDataset configuration
    dataset_cfg = create_ng_data_configuration(
        data_type="array",
        axes=axes,
        patch_size=(16, 16) if "Z" not in axes else (8, 16, 16),
        batch_size=2,
        channels=channels,
    )

    # create lightning module
    model = N2VModule(algorithm_config=algorithm_config)

    # create data module
    data = CareamicsDataModule(
        data_config=dataset_cfg,
        train_data=train_array,
        val_data=val_array,
    )

    # create prediction writer callback params
    dirpath = tmp_path / "predictions"
    predict_writer = PredictionWriterCallback(dirpath=dirpath)

    # create trainer
    trainer = Trainer(
        max_epochs=1,
        default_root_dir=tmp_path,
        callbacks=[
            ModelCheckpoint(
                dirpath=tmp_path / "checkpoints",
                filename="test_lightning_api",
            ),
            predict_writer,
        ],
    )

    # train
    trainer.fit(model, datamodule=data)

    # predict
    predict_writer.set_writing_strategy(write_type="tiff", tiled=tiled)
    if tiled:
        pred_dataset_cfg = dataset_cfg.convert_mode(
            new_mode="predicting",
            new_data_type="tiff",
            new_batch_size=4,
            new_patch_size=(16, 16) if "Z" not in axes else (8, 16, 16),
            overlap_size=(4, 4) if "Z" not in axes else (4, 4, 4),
        )
    else:
        pred_dataset_cfg = dataset_cfg.convert_mode(
            new_mode="predicting",
            new_data_type="tiff",
            new_batch_size=4,
        )

    predict_data = CareamicsDataModule(
        data_config=pred_dataset_cfg,
        pred_data=train_dir,
    )

    # predict
    predicted = trainer.predict(model, datamodule=predict_data)
    predicted_images, _ = convert_prediction(predicted, tiled=tiled)

    # assert predicted file exists
    assert (dirpath / file_name).is_file()
    save_data = tifffile.imread(dirpath / file_name)

    # note: predicted image is a list
    np.testing.assert_array_equal(save_data, predicted_images[0], verbose=True)


@pytest.mark.mps_gh_fail
@pytest.mark.parametrize(
    "shape, axes, channels",
    [
        ((32, 32), "YX", None),
        ((3, 32, 32), "CYX", None),
        ((3, 32, 32), "CYX", [1]),
        ((3, 32, 32), "CYX", [0, 2]),
        ((16, 32, 32), "ZYX", None),
        ((5, 16, 32, 32), "SZYX", None),
        ((3, 16, 32, 32), "CZYX", None),
        ((3, 16, 32, 32), "CZYX", [1]),
        ((3, 16, 32, 32), "CZYX", [0, 2]),
        ((5, 3, 16, 32, 32), "SCZYX", None),
        ((5, 3, 16, 32, 32), "SCZYX", [1]),
        ((5, 3, 16, 32, 32), "SCZYX", [0, 2]),
    ],
)
def test_smoke_n2v_zarr(tmp_path, shape, axes, channels):
    rng = np.random.default_rng(42)

    # training data
    train_array = rng.integers(0, 255, shape).astype(np.float32)
    val_array = rng.integers(0, 255, shape).astype(np.float32)

    # write train array to tiff
    train_dir = tmp_path / "train"
    train_dir.mkdir()
    file_name = "image.zarr"
    g = zarr.open_group(train_dir / file_name, mode="w")
    arr = g.create_array("array", data=train_array, chunks="auto")

    if "C" in axes:
        len_channels = shape[axes.index("C")]
    else:
        len_channels = 1

    algorithm_config = N2VAlgorithm(
        model={
            "architecture": "UNet",
            "conv_dims": 2 if "Z" not in axes else 3,
            "in_channels": len(channels) if channels is not None else len_channels,
            "num_classes": len(channels) if channels is not None else len_channels,
        }
    )
    # create NGDataset configuration
    dataset_cfg = create_ng_data_configuration(
        data_type="array",
        axes=axes,
        patch_size=(16, 16) if "Z" not in axes else (8, 16, 16),
        batch_size=2,
        channels=channels,
    )

    # create lightning module
    model = N2VModule(algorithm_config=algorithm_config)

    # create data module
    data = CareamicsDataModule(
        data_config=dataset_cfg,
        train_data=train_array,
        val_data=val_array,
    )

    # create prediction writer callback params
    dirpath = tmp_path / "predictions"
    predict_writer = PredictionWriterCallback(dirpath=dirpath)

    # create trainer
    trainer = Trainer(
        max_epochs=1,
        default_root_dir=tmp_path,
        callbacks=[
            ModelCheckpoint(
                dirpath=tmp_path / "checkpoints",
                filename="test_lightning_api",
            ),
            predict_writer,
        ],
    )

    # train
    trainer.fit(model, datamodule=data)

    # predict
    predict_writer.set_writing_strategy(write_type="zarr", tiled=True)
    pred_dataset_cfg = dataset_cfg.convert_mode(
        new_mode="predicting",
        new_data_type="zarr",
        new_batch_size=4,
        new_patch_size=(16, 16) if "Z" not in axes else (8, 16, 16),
        overlap_size=(4, 4) if "Z" not in axes else (4, 4, 4),
        new_in_memory=False,
    )

    predict_data = CareamicsDataModule(
        data_config=pred_dataset_cfg,
        pred_data=[str(arr.store_path)],
    )

    # predict
    predicted = trainer.predict(model, datamodule=predict_data)
    predicted_images, _ = convert_prediction(predicted, tiled=True)

    # assert predicted file exists
    output_file = dirpath / f"{file_name[:-len('.zarr')]}_output.zarr"
    assert output_file.exists()
    save_data = zarr.open_array(output_file, path="array")

    predicted_img = predicted_images[0]

    # zarr writer removes singleton dims if not present in axes
    if "C" not in axes:
        predicted_img = predicted_img.squeeze(axis=1)
    if "S" not in axes:
        predicted_img = predicted_img.squeeze(axis=0)

    np.testing.assert_array_equal(save_data, predicted_img, verbose=True)



"""Filter using mean and standard deviation thresholds."""

from collections.abc import Sequence

import numpy as np
from tqdm import tqdm

from careamics.dataset_ng.image_stack_loader import load_arrays
from careamics.dataset_ng.patch_extractor import PatchExtractor
from careamics.dataset_ng.patch_filter.patch_filter_protocol import PatchFilterProtocol
from careamics.dataset_ng.patching_strategies import TilingStrategy


class MeanStdPatchFilter(PatchFilterProtocol):
    """
    Filter patches based on mean and standard deviation thresholds.

    Attributes
    ----------
    mean_threshold : float
        Threshold for the mean of the patch.
    std_threshold : float
        Threshold for the standard deviation of the patch.
    p : float
        Probability of applying the filter to a patch.
    rng : np.random.Generator
        Random number generator for stochastic filtering.
    """

    def __init__(
        self,
        mean_threshold: float,
        std_threshold: float | None = None,
        p: float = 1.0,
        seed: int | None = None,
    ) -> None:
        """
        Create a MeanStdPatchFilter.

        This filter removes patches whose mean and standard deviation are both below
        specified thresholds. The filtering is applied with a probability `p`, allowing
        for stochastic filtering.

        Parameters
        ----------
        mean_threshold : float
            Threshold for the mean of the patch.
        std_threshold : float | None, default=None
            Threshold for the standard deviation of the patch. If None, then no
            standard deviation filtering is applied.
        p : float, default=1
            Probability of applying the filter to a patch. Must be between 0 and 1.
        seed : int | None, default=None
            Seed for the random number generator for reproducibility.

        Raises
        ------
        ValueError
            If mean_threshold or std_threshold is negative.
        ValueError
            If std_threshold is negative.
        ValueError
            If p is not between 0 and 1.
        """

        if mean_threshold < 0:
            raise ValueError("Mean threshold must be non-negative.")
        if std_threshold is not None and std_threshold < 0:
            raise ValueError("Std threshold must be non-negative.")
        if not (0 <= p <= 1):
            raise ValueError("Probability p must be between 0 and 1.")

        self.mean_threshold = mean_threshold
        self.std_threshold = std_threshold

        self.p = p
        self.rng = np.random.default_rng(seed)

    def filter_out(self, patch: np.ndarray) -> bool:
        """
        Determine whether to filter out a patch based on mean and std thresholds.

        Parameters
        ----------
        patch : numpy.NDArray
            The image patch to evaluate.

        Returns
        -------
        bool
            True if the patch should be filtered out, False otherwise.
        """

        if self.rng.uniform(0, 1) < self.p:
            patch_mean = np.mean(patch)
            patch_std = np.std(patch)

            return (patch_mean < self.mean_threshold) or (
                self.std_threshold is not None and patch_std < self.std_threshold
            )
        return False

    @staticmethod
    def filter_map(image: np.ndarray, patch_size: Sequence[int]) -> np.ndarray:
        """
        Compute the mean and std map of an image.

        The mean and std are computed over non-overlapping patches. This method can be
        used to assess a useful threshold for the MeanStd filter.

        Parameters
        ----------
        image : numpy.NDArray
            The full image to evaluate.
        patch_size : Sequence[int]
            The size of the patches to consider.

        Returns
        -------
        np.ndarray
            Stacked mean and std maps of the image.

        Raises
        ------
        ValueError
            If the image is not 2D or 3D.

        Example
        -------
        The `filter_map` method can be used to assess useful thresholds for the
        MeanStd filter.
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> from careamics.dataset_ng.patch_filter import MeanStdPatchFilter
        >>> rng = np.random.default_rng(42)
        >>> image = rng.binomial(20, 0.1, (256, 256)).astype(np.float32)
        >>> image[64:192, 64:192] = rng.normal(50, 3, (128, 128))
        >>> image[96:160, 96:160] = rng.poisson(image[96:160, 96:160])
        >>> patch_size = (16, 16)
        >>> meanstd_map = MeanStdPatchFilter.filter_map(image, patch_size)
        >>> fig, ax = plt.subplots(3, 3, figsize=(10, 10)) # doctest: +SKIP
        >>> for i, mean_thresh in enumerate([48 + i for i in range(3)]):
        ...     for j, std_thresh in enumerate([5 + i for i in range(3)]):
        ...         ax[i, j].imshow(
        ...             (meanstd_map[0, ...] > mean_thresh)
        ...             & (meanstd_map[1, ...] > std_thresh),
        ...             cmap="gray", vmin=0, vmax=1
        ...         ) # doctest: +SKIP
        ...         ax[i, j].set_title(
        ...             f"Mean: {mean_thresh}, Std: {std_thresh}"
        ...         ) # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP
        """
        if len(image.shape) < 2 or len(image.shape) > 3:
            raise ValueError("Image must be 2D or 3D.")

        axes = "YX" if len(patch_size) == 2 else "ZYX"

        mean = np.zeros_like(image, dtype=float)
        std = np.zeros_like(image, dtype=float)

        image_stacks = load_arrays(source=[image], axes=axes)
        extractor = PatchExtractor(image_stacks)
        tiling = TilingStrategy(
            data_shapes=[(1, 1, *image.shape)],
            patch_size=patch_size,
            overlaps=(0,) * len(patch_size),  # no overlap
        )

        for idx in tqdm(range(tiling.n_patches), desc="Computing Mean/STD map"):
            patch_spec = tiling.get_patch_spec(idx)
            patch = extractor.extract_patch(
                data_idx=0,
                sample_idx=0,
                coords=patch_spec["coords"],
                patch_size=patch_size,
            )

            coordinates = tuple(
                slice(patch_spec["coords"][i], patch_spec["coords"][i] + p)
                for i, p in enumerate(patch_size)
            )
            mean[coordinates] = np.mean(patch)
            std[coordinates] = np.std(patch)

        return np.stack([mean, std], axis=0)

    @staticmethod
    def apply_filter(
        filter_map: np.ndarray,
        mean_threshold: float,
        std_threshold: float | None = None,
    ) -> np.ndarray:
        """
        Apply mean and std thresholds to a filter map.

        The filter map is the output of the `filter_map` method.

        Parameters
        ----------
        filter_map : np.ndarray
            Stacked mean and std maps of the image.
        mean_threshold : float
            Threshold for the mean of the patch.
        std_threshold : float | None, default=None
            Threshold for the standard deviation of the patch. If None, then no
            standard deviation filtering is applied.

        Returns
        -------
        np.ndarray
            A binary map where True indicates patches that pass the filter.
        """
        if std_threshold is not None:
            return (filter_map[0, ...] > mean_threshold) & (
                filter_map[1, ...] > std_threshold
            )
        return filter_map[0, ...] > mean_threshold



"""A protocol for patch filtering."""

from collections.abc import Sequence
from typing import Protocol

import numpy as np


class PatchFilterProtocol(Protocol):
    """
    An interface for implementing patch filtering strategies.
    """

    def filter_out(self, patch: np.ndarray) -> bool:
        """
        Determine whether to filter out a given patch.

        Parameters
        ----------
        patch : numpy.NDArray
            The image patch to evaluate.

        Returns
        -------
        bool
            True if the patch should be filtered out (excluded), False otherwise.
        """
        ...

    @staticmethod
    def filter_map(
        image: np.ndarray,
        patch_size: Sequence[int],
    ) -> np.ndarray:
        """
        Compute a filter map for the entire image based on the patch filtering criteria.

        Parameters
        ----------
        image : numpy.NDArray
            The full image to evaluate.
        patch_size : Sequence[int]
            The size of the patches to consider.

        Returns
        -------
        numpy.NDArray
            A map where each element is the .
        """
        ...



import numpy as np
import pytest

from careamics.dataset_ng.normalization.running_quantile import QuantileEstimator

QUANTILE_PAIRS = [(0.01, 0.99), (0.05, 0.95), (0.1, 0.9)]


@pytest.mark.parametrize("n_patches", [5, 10, 20])
@pytest.mark.parametrize("lower_q, upper_q", QUANTILE_PAIRS)
def test_histogram_accuracy(n_patches, lower_q, upper_q):
    np.random.seed(42)
    estimator = QuantileEstimator(
        lower_quantiles=[lower_q],
        upper_quantiles=[upper_q],
    )
    all_data = []
    for _ in range(n_patches):
        patch = np.random.rand(1, 64, 64).astype(np.float32) * 65000
        estimator.update(patch)
        all_data.append(patch)

    lower, upper = estimator.finalize()
    all_data = np.concatenate(all_data, axis=1)

    assert np.isclose(lower[0], np.quantile(all_data, lower_q), rtol=0.01)
    assert np.isclose(upper[0], np.quantile(all_data, upper_q), rtol=0.01)


@pytest.mark.parametrize("lower_q, upper_q", QUANTILE_PAIRS)
def test_microscopy_like_distribution(lower_q, upper_q):
    np.random.seed(42)
    estimator = QuantileEstimator(lower_quantiles=[lower_q], upper_quantiles=[upper_q])

    all_patches = []
    for _ in range(50):
        patch = np.random.exponential(scale=50, size=(1, 64, 64)).astype(np.float32)
        n_spots = np.random.randint(1, 5)
        for _ in range(n_spots):
            y, x = np.random.randint(0, 64, 2)
            patch[0, y, x] = np.random.uniform(500, 1000)
        estimator.update(patch)
        all_patches.append(patch)

    lower, upper = estimator.finalize()
    all_data = np.concatenate(all_patches, axis=1)

    assert np.isclose(lower[0], np.quantile(all_data, lower_q), rtol=0.05)
    assert np.isclose(upper[0], np.quantile(all_data, upper_q), rtol=0.05)


@pytest.mark.parametrize("lower_q, upper_q", QUANTILE_PAIRS)
def test_bimodal_distribution(lower_q, upper_q):
    np.random.seed(42)
    estimator = QuantileEstimator(lower_quantiles=[lower_q], upper_quantiles=[upper_q])

    all_patches = []
    for _ in range(30):
        background = np.random.normal(loc=100, scale=10, size=(1, 32, 32))
        foreground = np.random.normal(loc=500, scale=50, size=(1, 32, 32))
        mask = np.random.random((1, 32, 32)) > 0.7
        patch = np.where(mask, foreground, background).astype(np.float32)
        estimator.update(patch)
        all_patches.append(patch)

    lower, upper = estimator.finalize()
    all_data = np.concatenate(all_patches, axis=1)

    assert np.isclose(lower[0], np.quantile(all_data, lower_q), rtol=0.05)
    assert np.isclose(upper[0], np.quantile(all_data, upper_q), rtol=0.05)


@pytest.mark.parametrize("lower_q, upper_q", QUANTILE_PAIRS)
def test_16bit_dynamic_range(lower_q, upper_q):
    np.random.seed(42)
    estimator = QuantileEstimator(lower_quantiles=[lower_q], upper_quantiles=[upper_q])

    all_patches = []
    for _ in range(20):
        patch = np.random.randint(0, 65535, size=(1, 64, 64)).astype(np.float32)
        estimator.update(patch)
        all_patches.append(patch)

    lower, upper = estimator.finalize()
    all_data = np.concatenate(all_patches, axis=1)

    assert np.isclose(lower[0], np.quantile(all_data, lower_q), atol=100)
    assert np.isclose(upper[0], np.quantile(all_data, upper_q), atol=100)


def test_multichannel_different_quantiles():
    np.random.seed(42)
    estimator = QuantileEstimator(
        lower_quantiles=[0.01, 0.05, 0.1],
        upper_quantiles=[0.99, 0.95, 0.9],
    )

    all_patches = []
    for _ in range(25):
        patch = np.random.rand(3, 32, 32).astype(np.float32) * 1000
        estimator.update(patch)
        all_patches.append(patch)

    lower, upper = estimator.finalize()
    all_data = np.concatenate(all_patches, axis=1)

    for ch, (lq, uq) in enumerate(
        zip([0.01, 0.05, 0.1], [0.99, 0.95, 0.9], strict=True)
    ):
        assert np.isclose(lower[ch], np.quantile(all_data[ch], lq), rtol=0.05)
        assert np.isclose(upper[ch], np.quantile(all_data[ch], uq), rtol=0.05)


@pytest.mark.parametrize("lower_q, upper_q", QUANTILE_PAIRS)
def test_rebinning_with_expanding_range(lower_q, upper_q):
    np.random.seed(42)
    estimator = QuantileEstimator(
        lower_quantiles=[lower_q],
        upper_quantiles=[upper_q],
    )

    all_patches = []
    for _ in range(10):
        patch = np.random.rand(1, 32, 32).astype(np.float32) * 100
        estimator.update(patch)
        all_patches.append(patch)

    for _ in range(10):
        patch = np.random.rand(1, 32, 32).astype(np.float32) * 1000
        estimator.update(patch)
        all_patches.append(patch)

    lower, upper = estimator.finalize()
    all_data = np.concatenate(all_patches, axis=1)

    assert np.isclose(lower[0], np.quantile(all_data, lower_q), rtol=0.1)
    assert np.isclose(upper[0], np.quantile(all_data, upper_q), rtol=0.1)


@pytest.mark.parametrize("constant_value", [0.0, 42.0, -100.0, 65535.0])
@pytest.mark.parametrize("lower_q, upper_q", QUANTILE_PAIRS)
def test_constant_data(constant_value, lower_q, upper_q):
    estimator = QuantileEstimator(lower_quantiles=[lower_q], upper_quantiles=[upper_q])
    data = np.full((1, 32, 32), constant_value, dtype=np.float32)

    estimator.update(data)
    lower, upper = estimator.finalize()

    assert np.isclose(lower[0], constant_value, atol=1e-5)
    assert np.isclose(upper[0], constant_value, atol=1e-5)


def test_no_data_returns_zero():
    estimator = QuantileEstimator(lower_quantiles=[0.1], upper_quantiles=[0.9])
    lower, upper = estimator.finalize()

    assert lower[0] == 0.0
    assert upper[0] == 0.0



"""Pydantic model for the max patch filter."""

from typing import Literal

from .filter_config import FilterConfig


class MaxFilterConfig(FilterConfig):
    """Pydantic model for the max patch filter."""

    name: Literal["max"] = "max"
    """Name of the filter."""

    threshold: float
    """Threshold for the minimum of the max-filtered patch."""



"""Filter patch using a maximum filter."""

from collections.abc import Sequence

import numpy as np
from scipy.ndimage import maximum_filter
from tqdm import tqdm

from careamics.dataset_ng.image_stack_loader import load_arrays
from careamics.dataset_ng.patch_extractor import PatchExtractor
from careamics.dataset_ng.patch_filter.patch_filter_protocol import PatchFilterProtocol
from careamics.dataset_ng.patching_strategies import TilingStrategy
from careamics.utils import get_logger

logger = get_logger(__name__)


class MaxPatchFilter(PatchFilterProtocol):
    """
    A patch filter based on thresholding the maximum filter of the patch.

    Inspired by the CSBDeep approach.

    Attributes
    ----------
    threshold : float
        Threshold for the maximum filter of the patch.
    p : float
        Probability of applying the filter to a patch.
    rng : np.random.Generator
        Random number generator for stochastic filtering.
    """

    def __init__(
        self,
        threshold: float,
        p: float = 1.0,
        threshold_ratio: float = 0.25,
        seed: int | None = None,
    ) -> None:
        """
        Create a MaxPatchFilter.

        This filter removes patches whose maximum filter valuepixels are below a
        specified threshold.

        Parameters
        ----------
        threshold : float
            Threshold for the maximum filter of the patch.
        p : float, default=1
            Probability of applying the filter to a patch. Must be between 0 and 1.
        threshold_ratio : float, default=0.25
            Ratio of pixels that must be below threshold for patch to be filtered out.
            Must be between 0 and 1.
        seed : int | None, default=None
            Seed for the random number generator for reproducibility.
        """
        self.threshold = threshold
        self.threshold_ratio = threshold_ratio
        self.p = p
        self.rng = np.random.default_rng(seed)

    def filter_out(self, patch: np.ndarray) -> bool:
        if self.rng.uniform(0, 1) < self.p:

            if np.max(patch) < self.threshold:
                return True

            patch_shape = [(p // 2 if p > 1 else 1) for p in patch.shape]
            filtered = maximum_filter(patch, patch_shape, mode="constant")
            return np.mean(filtered < self.threshold) > self.threshold_ratio

        return False

    @staticmethod
    def filter_map(
        image: np.ndarray,
        patch_size: Sequence[int],
    ) -> np.ndarray:
        """
        Compute the maximum map of an image.

        The map is computed over non-overlapping patches. This method can be used
        to assess a useful threshold for the MaxPatchFilter filter.

        Parameters
        ----------
        image : numpy.NDArray
            The image for which to compute the map, must be 2D or 3D.
        patch_size : Sequence[int]
            The size of the patches to compute the map over. Must be a sequence
            of two integers.

        Returns
        -------
        numpy.NDArray
            The max map of the patch.

        Raises
        ------
        ValueError
            If the image is not 2D or 3D.

        Example
        -------
        The `filter_map` method can be used to assess a useful threshold for the
        Shannon entropy filter. Below is an example of how to compute and visualize
        the Shannon entropy map of a random image and visualize thresholded versions
        of the map.
        >>> import numpy as np
        >>> from matplotlib import pyplot as plt
        >>> from careamics.dataset_ng.patch_filter import MaxPatchFilter
        >>> rng = np.random.default_rng(42)
        >>> image = rng.binomial(20, 0.1, (256, 256)).astype(np.float32)
        >>> image[64:192, 64:192] += rng.normal(50, 5, (128, 128))
        >>> image[96:160, 96:160] = rng.poisson(image[96:160, 96:160])
        >>> patch_size = (16, 16)
        >>> max_filtered = MaxPatchFilter.filter_map(image, patch_size)
        >>> fig, ax = plt.subplots(1, 5, figsize=(20, 5)) # doctest: +SKIP
        >>> for i, thresh in enumerate([50 + i*5 for i in range(5)]):
        ...     ax[i].imshow(max_filtered >= thresh, cmap="gray") # doctest: +SKIP
        ...     ax[i].set_title(f"Threshold: {thresh}") # doctest: +SKIP
        >>> plt.show() # doctest: +SKIP
        """
        if len(image.shape) < 2 or len(image.shape) > 3:
            raise ValueError("Image must be 2D or 3D.")

        axes = "YX" if len(patch_size) == 2 else "ZYX"

        max_filtered = np.zeros_like(image, dtype=float)

        image_stacks = load_arrays(source=[image], axes=axes)
        extractor = PatchExtractor(image_stacks)
        tiling = TilingStrategy(
            data_shapes=[(1, 1, *image.shape)],
            patch_size=patch_size,
            overlaps=(0,) * len(patch_size),  # no overlap
        )
        max_patch_size = [p // 2 for p in patch_size]

        for idx in tqdm(range(tiling.n_patches), desc="Computing max map"):
            patch_spec = tiling.get_patch_spec(idx)
            patch = extractor.extract_patch(
                data_idx=0,
                sample_idx=0,
                coords=patch_spec["coords"],
                patch_size=patch_size,
            )

            coordinates = tuple(
                slice(patch_spec["coords"][i], patch_spec["coords"][i] + p)
                for i, p in enumerate(patch_size)
            )
            max_filtered[coordinates] = maximum_filter(
                patch.squeeze(), max_patch_size, mode="constant"
            )

        return max_filtered

    @staticmethod
    def apply_filter(
        filter_map: np.ndarray,
        threshold: float,
    ) -> np.ndarray:
        """
        Apply the max filter to a filter map.

        The filter map is the output of the `filter_map` method.

        Parameters
        ----------
        filter_map : numpy.NDArray
            The max filter map of the image.
        threshold : float
            The threshold to apply to the filter map.

        Returns
        -------
        numpy.NDArray
            A boolean array where True indicates that the patch should be kept
            (not filtered out) and False indicates that the patch should be filtered
            out.
        """
        threshold_map = filter_map >= threshold
        coverage = np.sum(threshold_map) * 100 / threshold_map.size
        logger.info(f"Image coverage: {coverage:.2f}%")
        return threshold_map



"""Callback for stopping prediction based on external condition."""

from collections.abc import Callable
from typing import Any

from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import Callback


class PredictionStoppedException(Exception):
    """Exception raised when prediction is stopped by external signal."""

    pass


class StopPredictionCallback(Callback):
    """PyTorch Lightning callback to stop prediction based on external condition.

    This callback monitors a user-provided stop condition at the start of each
    prediction batch. When the condition is met, the callback stops the trainer
    and raises PredictionStoppedException to interrupt the prediction loop.

    Parameters
    ----------
    stop_condition : Callable[[], bool]
        A callable that returns True when prediction should stop. The callable
        is invoked at the start of each prediction batch.
    """

    def __init__(self, stop_condition: Callable[[], bool]) -> None:
        """Initialize the callback with a stop condition.

        Parameters
        ----------
        stop_condition : Callable[[], bool]
            Function that returns True when prediction should stop.
        """
        super().__init__()
        self.stop_condition = stop_condition

    def on_predict_batch_start(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int = 0,
    ) -> None:
        """Check stop condition at the start of each prediction batch.

        Parameters
        ----------
        trainer : Trainer
            PyTorch Lightning trainer instance.
        pl_module : LightningModule
            Lightning module being used for prediction.
        batch : Any
            Current batch of data.
        batch_idx : int
            Index of the current batch.
        dataloader_idx : int, optional
            Index of the dataloader, by default 0.

        Raises
        ------
        PredictionStoppedException
            If stop_condition() returns True.
        """
        if self.stop_condition():
            trainer.should_stop = True
            raise PredictionStoppedException("Prediction stopped by user")



"""Sequential patching Pydantic model."""

from typing import Literal

from ._overlapping_patched_config import _OverlappingPatchedConfig


class SequentialPatchingConfig(_OverlappingPatchedConfig):
    """Sequential patching Pydantic model.

    Attributes
    ----------
    name : "sequential"
        The name of the patching strategy.
    patch_size : sequence of int
        The size of the patch in each spatial dimension, each patch size must be a power
        of 2 and larger than 8.
    overlaps : list of int, optional
        The overlaps between patches in each spatial dimension. If `None`, no overlap is
        applied. The overlaps must be smaller than the patch size in each spatial
        dimension, and the number of dimensions be either 2 or 3.
    """

    name: Literal["sequential"] = "sequential"
    """The name of the patching strategy."""



"""Patch filtering strategies."""

__all__ = [
    "CoordinateFilterProtocol",
    "MaskCoordFilter",
    "MaxPatchFilter",
    "MeanStdPatchFilter",
    "PatchFilterProtocol",
    "ShannonPatchFilter",
    "create_coord_filter",
    "create_patch_filter",
]

from .coordinate_filter_protocol import CoordinateFilterProtocol
from .filter_factory import create_coord_filter, create_patch_filter
from .mask_filter import MaskCoordFilter
from .max_filter import MaxPatchFilter
from .mean_std_filter import MeanStdPatchFilter
from .patch_filter_protocol import PatchFilterProtocol
from .shannon_filter import ShannonPatchFilter
